import logging

import dns
from django.db import models
from django.conf import settings
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from munch.apps.users.models import Organization
from munch.core.mail.utils import extract_domain
from munch.core.utils.models import AbstractOwnedModel
from munch.core.utils.managers import OwnedModelQuerySet
from munch.core.utils.validators import DomainNameValidator
from .fields import DomainCheckField

log = logging.getLogger('munch')


class SendingDomainQueryset(OwnedModelQuerySet):
    def get_from_email_addr(self, email_addr, must_raise=True):
        domain = extract_domain(email_addr)
        if must_raise:
            return self.get(name=domain)
        return self.filter(name=domain).first()


class SendingDomain(AbstractOwnedModel):
    name = models.CharField(
        verbose_name=_('name'), max_length=200,
        validators=[DomainNameValidator()])
    creation_date = models.DateTimeField(default=timezone.now)
    update_date = models.DateTimeField(blank=True)
    organization = models.ForeignKey(
        Organization, related_name='domains', verbose_name=_('organization'))
    alt_organizations = models.ManyToManyField(Organization, blank=True)
    dkim_status = DomainCheckField(verbose_name=_('DKIM status'))
    dkim_status_date = models.DateTimeField(
        verbose_name=_('DKIM status last change'), blank=True, null=True)

    app_domain = models.CharField(
        _('Custom application domain'), max_length=200,
        blank=True, default='', help_text=_(
            'Domain to use for email links generation'),
        validators=[DomainNameValidator()])
    app_domain_status = DomainCheckField(
        verbose_name=_('Custom application domain status'))
    app_domain_status_date = models.DateTimeField(
        verbose_name=_('Custom application domain status last change'),
        blank=True, null=True)

    objects = SendingDomainQueryset.as_manager()

    owner_path = 'organization'
    author_path = AbstractOwnedModel.IRRELEVANT

    class Meta(AbstractOwnedModel.Meta):
        unique_together = (('name', 'organization'),)
        verbose_name = _("sending domain")
        verbose_name_plural = _("sending domains")

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        from .tasks import validate_sending_domain_field

        created = not self.pk
        if not created:
            previous = SendingDomain.objects.get(pk=self.pk)

        # Set changed fields to "Pending" or "Unknown"
        if created or previous.name != self.name:
            self.dkim_status = DomainCheckField.PENDING

        if self.app_domain:
            if created or previous.app_domain != self.app_domain:
                self.app_domain_status = DomainCheckField.PENDING
        else:
            if not created and previous.app_domain != self.app_domain:
                self.app_domain_status = DomainCheckField.UNKNOWN

        self.update_date = timezone.now()

        super().save(*args, **kwargs)

        if created:
            validate_sending_domain_field.apply_async([self.id, 'dkim'])
            if self.app_domain:
                validate_sending_domain_field.apply_async(
                    [self.id, 'app_domain'])
        else:
            if self.app_domain and (previous.app_domain != self.app_domain):
                validate_sending_domain_field.apply_async(
                    [self.id, 'app_domain'])

    def validate_app_domain(self):
        if self.app_domain:
            log.debug('Validating SendingDomain app domain {} (pk:{})'.format(
                self.app_domain, self.pk))
            try:
                # a CNAME record should be unique
                # we also have to strip the trailing dot
                res = dns.resolver.query(
                    self.app_domain, 'CNAME')[0].to_text().strip('.')
            except (IndexError, dns.exception.DNSException):
                log.debug(
                    'No DNS answer for SendingDomain app '
                    'domain {} (pk:{})'.format(self.app_domain, self.pk))
                self.app_domain_status = DomainCheckField.NOT_CONFIGURED
            else:
                if (res == settings.USERS['ORGANIZATION_APP_DOMAIN_CNAME']):
                    log.debug(
                        'SendingDomain app domain {} correctly '
                        'configured (pk:{})'.format(self.app_domain, self.pk))
                    self.app_domain_status = DomainCheckField.OK
                else:
                    log.debug(
                        'SendingDomain app domain {} badly '
                        'configured (pk:{}) -> DNS record value: '.format(
                            self.app_domain, self.pk, res))
                    self.app_domain_status = DomainCheckField.BADLY_CONFIGURED

        return self.app_domain_status

    def validate_dkim(self):
        keydomain = '{}._domainkey.{}'.format(
            settings.DOMAINS['DKIM_KEY_ID'], self.name)
        try:
            # a CNAME record should be unique
            # we also have to strip the trailing dot
            res = dns.resolver.query(
                keydomain, 'CNAME')[0].to_text().strip('.')
        except (IndexError, dns.exception.DNSException):
            #
            # TODO: remove this when all current domains are migrated to CNAME
            # and simply fail with NOT_CONFIGURED
            #
            # Try to look for a standard TXT record
            try:
                # its not supposed to have several TXT records with same key
                res = dns.resolver.query(keydomain, 'TXT')[0].to_text()
            except (IndexError, dns.exception.DNSException):
                log.debug(
                    'No DNS answer for SendingDomain DKIM '
                    'record {} (pk:{})'.format(keydomain, self.pk))
                self.dkim_status = DomainCheckField.NOT_CONFIGURED
            else:
                if res == settings.DOMAINS['DKIM_KEY_CONTENT']:
                    log.debug(
                        'SendingDomain DKIM record {} correctly '
                        'configured (pk:{})'.format(keydomain, self.pk))
                    self.dkim_status = DomainCheckField.OK
                else:
                    log.debug(
                        'SendingDomain DKIM record {} badly '
                        'configured (pk:{}) -> DNS record value: {}'.format(
                            keydomain, self.pk, res))
                    self.dkim_status = DomainCheckField.BADLY_CONFIGURED
        else:
            if res == settings.DOMAINS['DKIM_CNAME']:
                log.debug(
                    'SendingDomain DKIM record {} correctly '
                    'configured (pk:{})'.format(keydomain, self.pk))
                self.dkim_status = DomainCheckField.OK
            else:
                log.debug(
                    'SendingDomain DKIM record {} badly '
                    'configured (pk:{}) -> DNS record value: {}'.format(
                        keydomain, self.pk, res))
                self.dkim_status = DomainCheckField.BADLY_CONFIGURED

        return self.dkim_status

    def is_valid(self):
        if self.dkim_status != DomainCheckField.OK:
            return False
        if self.app_domain and self.app_domain_status != DomainCheckField.OK:
            return False
        return True



from django.conf.urls import url
from django.conf.urls import include

from ...utils.routers import APIRouter
from munch.urls import api_urlpatterns_v1

from . import views

router = APIRouter()
router.register('categories', views.CategoryViewSet, base_name='category')

api_urlpatterns_v1 += [url('', include(router.urls, namespace='core'))]



import copy

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.routers import SimpleRouter
from rest_framework.routers import DefaultRouter


class APIRouter(DefaultRouter):
    routes = copy.deepcopy(SimpleRouter.routes)
    routes[0].mapping.update({
        'put': 'bulk_update',
        'patch': 'partial_bulk_update',
        'delete': 'bulk_destroy'})

    def get_api_root_view(self, api_urls):
        class APIRoot(APIView):
            def get(self, request, format=None):
                return Response({})
        return APIRoot.as_view()



from django.urls import reverse
from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.db import transaction
from django.utils.translation import ugettext_lazy as _

from munch.apps.domains.admin import SendingDomainInline

from .models import MunchUser
from .models import Organization
from .models import APIApplication
from .models import SmtpApplication
from .models import OrganizationSettings
from .forms import MunchUserCreationForm


class SmtpApplicationAdmin(admin.ModelAdmin):
    list_display = ['identifier', 'username', 'secret', 'author_url']
    actions = ['regenerate_credentials']
    readonly_fields = ['username', 'secret']
    search_fields = ['identifier', 'username', 'secret', 'author__identifier']
    raw_id_fields = ['author']

    def regenerate_credentials(self, request, queryset):
        with transaction.atomic():
            for application in queryset:
                application.regen_credentials()
                application.save()

    def author_url(self, obj):
        return '<a href="{}">{}</a>'.format(
            reverse('admin:users_munchuser_change', args=[obj.author.pk]),
            obj.author.identifier)
    author_url.allow_tags = True


class SmtpApplicationInline(admin.TabularInline):
    model = SmtpApplication
    extra = 0
    readonly_fields = ['username', 'secret']


class APIApplicationInline(admin.TabularInline):
    model = APIApplication
    extra = 0
    readonly_fields = ['secret']


class APIApplicationAdmin(admin.ModelAdmin):
    list_display = ['identifier', 'secret', 'author_url']
    actions = ['regenerate_secret']
    readonly_fields = ['secret']
    search_fields = ['identifier', 'secret', 'author__identifier']
    raw_id_fields = ['author']

    def regenerate_secret(self, request, queryset):
        with transaction.atomic():
            for application in queryset:
                application.regen_secret()
                application.save()

    def author_url(self, obj):
        return '<a href="{}">{}</a>'.format(
            reverse('admin:users_munchuser_change', args=[obj.author.pk]),
            obj.author.identifier)
    author_url.allow_tags = True


class OrganizationInline(admin.TabularInline):
    model = Organization
    extra = 0
    readonly_fields = ['creation_date', 'update_date']


class OrganizationAdmin(admin.ModelAdmin):
    model = Organization
    search_fields = ['name', 'contact_email']
    readonly_fields = ['creation_date', 'update_date']
    list_display = [
        'name', 'contact_email', 'parent', 'creation_date', 'update_date']
    list_filter = [
        'can_external_optout', 'can_attach_files',
        'creation_date', 'update_date']
    raw_id_fields = ['parent']
    inlines = [OrganizationInline, SendingDomainInline]


class OrganizationSettingsAdmin(admin.ModelAdmin):
    list_display = ['organization', 'nickname']
    search_fields = ['organization__name']


class MunchUserAdmin(UserAdmin):
    """
    Presents both human and apps, so its a bit rough for now.

    Room for improvement would be to present a distinction similar to the one
    made in API.
    """
    actions = ['reset_password']
    list_display = [
        'identifier', 'full_name', 'organization', 'last_login']
    list_filter = ['is_active', 'is_admin']
    search_fields = ['identifier', 'organization__name']
    fieldsets = None
    add_form = MunchUserCreationForm
    raw_id_fields = ['organization', 'invited_by']

    readonly_fields = ['secret', 'last_login']
    ordering = ['identifier']
    inlines = [SmtpApplicationInline, APIApplicationInline]

    def get_fieldsets(self, request, obj=None):
        return super(UserAdmin, self).get_fieldsets(request, obj)

    def reset_password(self, request, queryset):
        user_qs = MunchUser.objects.filter(
            pk__in=queryset.values_list('pk', flat=True))

        for user in user_qs:
            user.send_password_reset_email()
        self.message_user(
            request,
            _('{} password reset email(s) sent').format(
                user_qs.count()))
    reset_password.short_description = _('Reset password (via email)')


MunchUserAdmin.add_fieldsets = MunchUserAdmin.fieldsets


admin.site.register(MunchUser, MunchUserAdmin)
admin.site.register(Organization, OrganizationAdmin)
admin.site.register(APIApplication, APIApplicationAdmin)
admin.site.register(SmtpApplication, SmtpApplicationAdmin)
admin.site.register(OrganizationSettings, OrganizationSettingsAdmin)



from unittest.mock import patch

from django.test import TestCase
from django.conf import settings
from django.core.files.base import ContentFile
from django.test.utils import override_settings
from django.core.mail import EmailMultiAlternatives
from django.core.files.storage import default_storage
from slimta.envelope import Envelope

from munch.core.utils.tests import temporary_settings
from munch.core.mail.models import return_path_parser
from munch.core.mail.models import unsubscribe_parser
from munch.apps.users.tests.factories import UserFactory
from munch.apps.spamcheck.tests import get_spam_result_mock
from munch.apps.optouts.tests.factories import OptOutFactory

from ..tasks import handle_dsn
from ..models import Mail
from ..models import Message
from ..models import MailStatus
from ..models import MessageAttachment

from .factories import MailFactory
from .factories import MessageFactory
from .factories import MailStatusFactory


class MailTest(TestCase):
    @override_settings(BYPASS_DNS_CHECKS=True)
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

    def test_new_mail_creates_mailstatus(self):
        statuses = MailStatus.objects.filter(mail__message=self.message)
        self.assertEqual(statuses.count(), 0)
        mail = MailFactory(message=self.message)
        self.assertEqual(mail.statuses.count(), 1)
        self.assertEqual(mail.statuses.first().status, MailStatus.UNKNOWN)

    def test_new_bulk_mail_creates_mailstatus(self):
        statuses = MailStatus.objects.filter(mail__message=self.message)
        self.assertEqual(statuses.count(), 0)
        mail_attrs = MailFactory.attributes()
        mail_attrs.update({'message': self.message})
        Mail.objects.bulk_create([Mail(**mail_attrs)])
        m = Mail.objects.get(recipient=mail_attrs['recipient'])
        self.assertEqual(m.statuses.count(), 1)
        self.assertEqual(m.statuses.first().status, MailStatus.UNKNOWN)
        self.assertEqual(m.statuses.last().creation_date, m.latest_status_date)
        self.assertEqual(m.statuses.first().creation_date, m.first_status_date)

    def test_new_mailstatus_updates_mail(self):
        mail = MailFactory(message=self.message)
        mail_status = MailStatusFactory(mail=mail, status=MailStatus.QUEUED)
        statuses = MailStatus.objects.filter(mail__message=self.message)
        self.assertEqual(statuses.count(), 2)
        self.assertEqual(mail.statuses.last(), mail_status)
        self.assertEqual(mail.curstatus, MailStatus.QUEUED)

    def test_to_mail(self):
        mail = MailFactory(message=self.message)
        message = mail.as_message()
        self.assertIsInstance(message, EmailMultiAlternatives)

    def test_to_mail_includes_unsubscription_link(self):
        mail = MailFactory(message=self.message)
        self.assertIn('/optout/', mail.as_message().body)

    def test_to_mail_html_includes_view_in_browser(self):
        user = UserFactory()
        message = MessageFactory(
            author=user, html='WEB_VERSION_URL UNSUBSCRIBE_URL')
        mail = MailFactory(message=message)
        self.assertIn('/archive/', mail.as_message().body)

    def test_to_mail_html_does_not_include_view_in_browser(self):
        user = UserFactory()
        message = MessageFactory(author=user, html='test UNSUBSCRIBE_URL')
        mail = MailFactory(message=message)
        self.assertNotIn('/archive/', mail.as_message().body)

    def test_to_mail_includes_ok_topic(self):
        mail = MailFactory(message=self.message)
        message = mail.as_message()
        self.assertEqual(
            message.message().get('Subject'), self.message.subject)

    def test_addrs(self):
        mail = MailFactory(message=self.message)
        self.assertTrue(unsubscribe_parser.is_valid(mail.unsubscribe_addr))
        self.assertTrue(return_path_parser.is_valid(mail.envelope_from))


class MailMergeTests(TestCase):
    def test_mailmerge(self):
        user = UserFactory()
        message = MessageFactory(
            author=user, html='<h1>Hi {{ first_name }} %s</h1>' % (
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER']))
        mail = MailFactory(message=message, properties={'first_name': 'John'})

        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            message.save()

        content = message.to_mail(mail)
        self.assertNotIn('{{ first_name }}', content.alternatives[0][0])
        self.assertIn('John', content.alternatives[0][0])


@override_settings(OPTOUTS_ADDRESS='')
class MailSending(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

        settings.CAMPAIGNS['SKIP_SPAM_CHECK'] = True

    def test_fire_message(self):
        MailFactory(message=self.message)

        a_mail = MailFactory(message=self.message)
        self.assertNotEqual(a_mail.identifier, '')

        self.message.status = 'sending'
        self.message.author.organization.settings.notify_message_status = False
        self.message.save()  # Fires the message

        self.assertEqual(self.message.mails.filter(
            curstatus=MailStatus.SENDING).count(), 2)
        self.assertEqual(self.message.mails.filter(
            curstatus=MailStatus.DELIVERED).count(), 0)
        for mail in self.message.mails.all():
            MailStatusFactory(mail=mail, status=MailStatus.DELIVERED)
        self.assertEqual(self.message.status, self.message.SENT)

    def test_fire_message_handles_optouted(self):
        """
        Checks that an ignored recipient for cause of previous optout on its
        address will  put the Mail in IGNORED state, and will not prevent the
        other mail to reach its SENDING state and the Message to complete.
        """
        mail = MailFactory(message=self.message)
        OptOutFactory(
            author=self.message.author, category=self.message.category,
            identifier=mail.identifier, address=mail.recipient)

        # A prev message with same recipient had an optout
        message = MessageFactory(author=self.user)
        mail_ok = MailFactory(message=message)
        mail_optouted = MailFactory(message=message, recipient=mail.recipient)

        message.status = 'sending'
        message.author.organization.settings.notify_message_status = False
        message.save()  # Fires the message

        self.assertEqual(message.mails.filter(
            curstatus=MailStatus.SENDING).count(), 1)
        self.assertEqual(message.mails.filter(
            curstatus=MailStatus.DELIVERED).count(), 0)
        self.assertEqual(Mail.objects.get(
            pk=mail_optouted.pk).curstatus, MailStatus.IGNORED)

        handle_dsn(
            'To: {}'.format(mail_ok.envelope_from),
            {
                'Final-Recipient': mail_ok.recipient,
                'Diagnostic-Code': 'Delivered',
                'Status': '2.0.0',
                'Arrival-Date': 'Fri,  5 Aug 2014 23:35:50 +0700 (WIT)'})

        self.assertEqual(
            Message.objects.get(pk=message.pk).status, message.SENT)

    def test_fire_message_good_headers(self):
        """ Some headers are required for mass mailing, check them """
        MailFactory(message=self.message)
        MailFactory(message=self.message)

        self.message.status = 'sending'
        self.message.author.organization.settings.notify_message_status = False
        self.message.save()

        # Fires the message
        self.assertEqual(self.message.mails.filter(
            curstatus=MailStatus.SENDING).count(), 2)
        self.assertEqual(self.message.mails.filter(
            curstatus=MailStatus.DELIVERED).count(), 0)
        for mail in self.message.mails.all():
            MailStatus.objects.create(mail=mail, status=MailStatus.DELIVERED)
        self.assertEqual(self.message.status, self.message.SENT)

        first_mail = self.message.mails.all().first()
        first_headers = dict(first_mail.as_message().message())
        self.assertIn('List-Unsubscribe', first_headers)
        self.assertEqual(first_headers['Precedence'], 'bulk')

        # checks that the url in the email headers for abuse is reachable
        abuse_msg = first_headers['X-Report-Abuse']
        abuse_url = abuse_msg.split(' ')[-1]
        response = self.client.get(abuse_url)
        self.assertEqual(response.status_code, 200)

    def test_fire_message_with_attachments(self):
        MailFactory(message=self.message)

        default_storage.save('foo.txt', ContentFile('testfoo'))

        ma = MessageAttachment(message=self.message, file='foo.txt')
        self.message.author.organization.settings.notify_message_status = False
        ma.save()

        # Fires the message
        self.message.status = 'sending'
        self.message.save()

        first_mail = self.message.mails.all().first()
        message = first_mail.message.to_mail(first_mail)
        self.assertEqual(len(message.attachments), 1)
        filename, content, mimetype = message.attachments[0]
        self.assertEqual(filename, 'foo.txt')
        self.assertEqual(content, 'testfoo')
        self.assertEqual(mimetype, 'text/plain')

        default_storage.delete('foo.txt')

    def test_fire_message_with_bin_attachments(self):
        MailFactory(message=self.message)

        default_storage.save(
            'foo.bmp',
            ContentFile(
                b'BM\x8e\x00\x00\x00\x00\x00\x00\x00\x8a\x00\x00\x00|\x00\x00'
                b'\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01\x00\x10\x00\x03\x00'
                b'\x00\x00\x04\x00\x00\x00\x13\x0b\x00\x00\x13\x0b\x00\x00\x00'
                b'\x00\x00\x00\x00\x00\x00\x00\x00\xf8\x00\x00\xe0\x07\x00\x00'
                b'\x1f\x00\x00\x00\x00\x00\x00\x00BGRs\x00\x00\x00\x00\x00\x00'
                b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00'
                b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff'
                b'\x00\x00'))

        ma = MessageAttachment(message=self.message, file='foo.bmp')
        ma.save()

        # Fires the message
        self.message.status = 'sending'
        self.message.author.organization.settings.notify_message_status = False
        self.message.save()

        first_mail = self.message.mails.all().first()
        message = first_mail.message.to_mail(first_mail)
        self.assertEqual(len(message.attachments), 1)
        filename, content, mimetype = message.attachments[0]
        self.assertEqual(filename, 'foo.bmp')
        self.assertEqual(len(content), 142)
        # the mimetype depends on python versions.
        self.assertIn(mimetype, ('application/octet-stream', 'image/x-ms-bmp'))

        default_storage.delete('foo.bmp')



# -*- coding: utf-8 -*-
from django.db import migrations
from django.core.management.sql import emit_post_migrate_signal

PERMISSIONS = {
    'contactlistpolicy': [
        ('add_contactlistpolicy', 'Can add contact list policy'),
        ('change_contactlistpolicy', 'Can change contact list policy'),
        ('delete_contactlistpolicy', 'Can delete contact list policy'), ],
    'contactqueuepolicyattribution': [
        ('add_contactqueuepolicyattribution', 'Can add contact queue policy attribution'),
        ('change_contactqueuepolicyattribution', 'Can change contact queue policy attribution'),
        ('delete_contactqueuepolicyattribution', 'Can delete contact queue policy attribution'), ],
    'contactlistpolicyattribution': [
        ('add_contactlistpolicyattribution', 'Can add contact list policy attribution'),
        ('change_contactlistpolicyattribution', 'Can change contact list policy attribution'),
        ('delete_contactlistpolicyattribution', 'Can delete contact list policy attribution'), ],
    'contact': [
        ('add_contact', 'Can add contact'),
        ('change_contact', 'Can change contact'),
        ('change_mine_contact', 'Can change mine contact'),
        ('change_organizations_contact', 'Can change organizations contact'),
        ('delete_contact', 'Can delete contact'),
        ('delete_mine_contact', 'Can delete mine contact'),
        ('delete_organizations_contact', 'Can delete organizations contact'),
        ('view_mine_contact', 'Can view mine contact'),
        ('view_organizations_contact', 'Can view organizations contact'), ],
    'collectedcontact': [
        ('add_collectedcontact', 'Can add collected address')
,        ('change_collectedcontact', 'Can change collected address'),
        ('delete_collectedcontact', 'Can delete collected address'), ],
    'contactlist': [
        ('add_contactlist', 'Can add liste de contacts'),
        ('change_contactlist', 'Can change liste de contacts'),
        ('change_mine_contactlist', 'Can change mine contactlist'),
        ('change_organizations_contactlist', 'Can change organizations contactlist'),
        ('delete_contactlist', 'Can delete liste de contacts'),
        ('delete_mine_contactlist', 'Can delete mine contactlist'),
        ('delete_organizations_contactlist', 'Can delete organizations contactlist'),
        ('view_mine_contactlist', 'Can view mine contactlist'),
        ('view_organizations_contactlist', 'Can view organizations contactlist'), ],
    'contactqueue': [
        ('add_contactqueue', 'Can add contact queue'),
        ('change_contactqueue', 'Can change contact queue'),
        ('change_organizations_contactqueue', 'Can change_organizations_contactqueue'),
        ('delete_contactqueue', 'Can delete contact queue'),
        ('delete_organizations_contactqueue', 'Can delete_organizations_contactqueue'),
        ('view_organizations_contactqueue', 'Can view_organizations_contactqueue'), ],
}
GROUP_PERMISSIONS = {
    'administrators': {
        'contactlistpolicy': [],
        'contactqueuepolicyattribution': [],
        'contactlistpolicyattribution': [],
        'contact': [
            'add_contact',
            'change_organizations_contact',
            'delete_organizations_contact',
            'view_organizations_contact', ],
        'collectedcontact': [],
        'contactlist': [
            'add_contactlist',
            'change_organizations_contactlist',
            'delete_contactlist',
            'delete_organizations_contactlist',
            'view_organizations_contactlist', ],
        'contactqueue': [
            'add_contactqueue',
            'change_organizations_contactqueue',
            'delete_organizations_contactqueue',
            'view_organizations_contactqueue', ],
    },
    'managers': {
        'contactlistpolicy': [],
        'contactqueuepolicyattribution': [],
        'contactlistpolicyattribution': [],
        'contact': [
            'add_contact',
            'change_organizations_contact',
            'delete_organizations_contact',
            'view_organizations_contact', ],
        'collectedcontact': [],
        'contactlist': [
            'add_contactlist',
            'change_organizations_contactlist',
            'delete_contactlist',
            'delete_organizations_contactlist',
            'view_organizations_contactlist', ],
        'contactqueue': [
            'add_contactqueue',
            'change_organizations_contactqueue',
            'delete_organizations_contactqueue',
            'view_organizations_contactqueue', ],
    },
    'users': {
        'contactlistpolicy': [],
        'contactqueuepolicyattribution': [],
        'contactlistpolicyattribution': [],
        'contact': [
            'add_contact',
            'change_mine_contact',
            'delete_mine_contact',
            'view_mine_contact', ],
        'collectedcontact': [],
        'contactlist': [
            'add_contactlist',
            'change_mine_contactlist',
            'delete_mine_contactlist',
            'view_mine_contactlist', ],
        'contactqueue': [],
    },
    'collaborators': {
        'contactlistpolicy': [],
        'contactqueuepolicyattribution': [],
        'contactlistpolicyattribution': [],
        'contact': [
            'add_contact',
            'change_organizations_contact',
            'delete_organizations_contact',
            'view_organizations_contact', ],
        'collectedcontact': [],
        'contactlist': [
            'add_contactlist',
            'change_organizations_contactlist',
            'delete_contactlist',
            'delete_organizations_contactlist',
            'view_organizations_contactlist', ],
        'contactqueue': [
            'add_contactqueue',
            'change_organizations_contactqueue',
            'delete_organizations_contactqueue',
            'view_organizations_contactqueue', ],
    },
}


def update_content_types(apps, schema_editor):
    db_alias = schema_editor.connection.alias
    emit_post_migrate_signal(False, 'default', db_alias)


def load_permissions(apps, schema_editor):
    Group = apps.get_model('auth', 'group')
    Permission = apps.get_model('auth', 'permission')
    ContentType = apps.get_model('contenttypes', 'contenttype')

    # Delete previous permissions
    for model in PERMISSIONS:
        content_type = ContentType.objects.get(
            app_label='contacts', model=model)
        Permission.objects.filter(content_type=content_type).delete()

    # Load permissions
    for model_name, permissions in PERMISSIONS.items():
        for permission_codename, permission_name in permissions:
            content_type = ContentType.objects.get(
                app_label='contacts', model=model_name)
            if not Permission.objects.filter(
                    codename=permission_codename,
                    content_type=content_type).exists():
                Permission.objects.create(
                    name=permission_name,
                    codename=permission_codename,
                    content_type=content_type)

    # Group permissions
    for group_name, models in GROUP_PERMISSIONS.items():
        group, _ = Group.objects.get_or_create(name=group_name)
        for model_name, permissions in models.items():
            content_type = ContentType.objects.get(
                app_label='contacts', model=model_name)
            for codename in permissions:
                permission, _ = Permission.objects.get_or_create(
                    codename=codename, content_type_id=content_type.id)
                group.permissions.add(permission)
                group.save()


class Migration(migrations.Migration):
    dependencies = [('contacts', '0001_initial')]
    operations = [
        migrations.RunPython(update_content_types, reverse_code=None),
        migrations.RunPython(load_permissions, reverse_code=None)]



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-22 10:18
from __future__ import unicode_literals

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import munch.core.mail.utils
import munch.core.models
import munch.core.utils.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0008_alter_user_username_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='MunchUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('identifier', models.EmailField(help_text='identifier', max_length=254, unique=True, verbose_name='identifier')),
                ('is_active', models.BooleanField(default=False, verbose_name='active')),
                ('is_admin', models.BooleanField(default=False, verbose_name='admin')),
                ('secret', models.CharField(max_length=30, verbose_name='secret')),
                ('creation_date', models.DateTimeField(auto_now_add=True, verbose_name='creation date')),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
            ],
            options={
                'verbose_name_plural': 'users',
                'verbose_name': 'user',
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations'),
            },
            bases=(munch.core.models.ValidationSignalsModel, models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='APIApplication',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(max_length=50, verbose_name='identifier')),
                ('secret', models.CharField(default=munch.core.mail.utils.mk_base64_uuid, max_length=25, verbose_name='secret')),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='author')),
            ],
            options={
                'verbose_name_plural': 'api applications',
                'verbose_name': 'api application',
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='Organization',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Organization name', max_length=200, verbose_name='name')),
                ('contact_email', models.EmailField(help_text='Main contact email for organization', max_length=254, verbose_name='contact email address')),
                ('can_external_optout', models.BooleanField(default=False, verbose_name='Can create messages with external optout')),
                ('can_attach_files', models.BooleanField(default=False, verbose_name='Can add attachments')),
                ('password_reset_link', models.CharField(blank=True, max_length=200, validators=[django.core.validators.RegexValidator('{{\\s+api_reset_link\\s+}}')], verbose_name='password reset link')),
                ('creation_date', models.DateTimeField(default=django.utils.timezone.now)),
                ('update_date', models.DateTimeField(blank=True)),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='users.Organization')),
            ],
            options={
                'verbose_name_plural': 'organizations',
                'verbose_name': 'organization',
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations'),
            },
            bases=(munch.core.models.ValidationSignalsModel, models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='OrganizationSettings',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('nickname', models.CharField(blank=True, default='', help_text='Organization nickname for tracking links', max_length=30, validators=[django.core.validators.RegexValidator('^[a-z0-9-]{4,30}$', message='Nickname can only contain lowercase letters, numbers and dashes. It must be between 4 and 30 characters long')], verbose_name='nickname')),
                ('notification_email', models.EmailField(blank=True, default='', help_text='Email which will receive optout and status update notifications', max_length=254, verbose_name='notification email')),
                ('notify_message_status', models.BooleanField(default=True, verbose_name='Receive message status notification emails')),
                ('notify_optouts', models.BooleanField(default=False, verbose_name='Receive optouts notification emails')),
                ('external_optout_message', models.TextField(blank=True, default='', help_text='Text which will be display to recipients who want to contact you in order to unsubscribe in case of external optout', verbose_name='External optout contact')),
                ('organization', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='settings', to='users.Organization', verbose_name='organization')),
            ],
            options={
                'verbose_name_plural': 'organization settings',
                'verbose_name': 'organization settings',
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations'),
                'abstract': False,
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='SmtpApplication',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(max_length=50, verbose_name='identifier')),
                ('username', models.CharField(db_index=True, default=munch.core.mail.utils.mk_base64_uuid, max_length=25, unique=True, verbose_name='username')),
                ('secret', models.CharField(default=munch.core.mail.utils.mk_base64_uuid, max_length=25, verbose_name='secret')),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='author')),
            ],
            options={
                'verbose_name_plural': 'smtp applications',
                'verbose_name': 'smtp application',
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.AddField(
            model_name='munchuser',
            name='organization',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='users', to='users.Organization', verbose_name='organization'),
        ),
        migrations.AddField(
            model_name='munchuser',
            name='user_permissions',
            field=models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions'),
        ),
        migrations.AlterUniqueTogether(
            name='smtpapplication',
            unique_together=set([('identifier', 'author')]),
        ),
        migrations.AlterUniqueTogether(
            name='apiapplication',
            unique_together=set([('identifier', 'author')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-24 17:41
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('users', '0003_auto_20160924_1312'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='organization',
            name='password_reset_link',
        ),
    ]



import re
import unittest

from django.core import mail
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.test import TestCase
from django.test.utils import override_settings

from ..models import MunchUser
from ..models import Organization

from .factories import UserFactory


class OrganizationTests(TestCase):
    def test_settings_auto_created(self):
        c = Organization(name='Test', contact_email='test@example.com')
        c.save()
        self.assertTrue(c.settings)


@override_settings(
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    CELERY_ALWAYS_EAGER=True, BROKER_BACKEND='memory')
class OrganizationSettingsTests(TestCase):
    def test_external_optout_create(self):
        c = Organization(
            name='Test', can_external_optout=True,
            contact_email='test@example.com')
        c.save()
        c.settings.external_optout_message = 'Jean-Jack : 01.01.01.01.01'
        c.settings.save()
        self.assertEqual(c.name, 'Test')
        self.assertEqual(
            c.settings.external_optout_message, 'Jean-Jack : 01.01.01.01.01')

    def test_external_optout_message_not_allowed(self):
        with self.assertRaises(ValidationError):
            c = Organization(
                name='Test', can_external_optout=False,
                contact_email='test@example.com')
            c.save()
            c.settings.external_optout_message = 'Jean-Jack : 01.01.01.01.01'
            c.settings.full_clean()
            c.settings.save()


class TestSecret(TestCase):
    def test_key_generation(self):
        s1 = MunchUser._mk_random_secret()
        self.assertEqual(len(s1), MunchUser.SECRET_LENGTH)

    def test_key_differs(self):
        s1 = MunchUser._mk_random_secret()
        s2 = MunchUser._mk_random_secret()
        self.assertNotEqual(s1, s2)

    def test_create_user_gens_secret(self):
        c = Organization.objects.filter(pk=1024).first()
        u = MunchUser.objects.create(
            identifier='foo@bar.example.com', organization=c,
            last_login=timezone.now())
        self.assertEqual(len(u.secret), MunchUser.SECRET_LENGTH)


@override_settings(
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    CELERY_ALWAYS_EAGER=True, BROKER_BACKEND='memory')
class TestMunchUser(TestCase):
    @unittest.skip('Must check if email backend successfully send email')
    def test_create_user(self):
        user = UserFactory(
            identifier='foo@example.com', last_login=timezone.now())
        self.assertEqual(user.identifier, 'foo@example.com')
        # account activation email
        self.assertEqual(len(mail.outbox), 0)


@override_settings(OPTOUTS_ADDRESS='http://testserver')
class TestPasswordReset(TestCase):
    @unittest.skip('Must check if email backend successfully send email')
    def test_password_reset(self):
        user = MunchUser.objects.create(
            is_active=True, is_admin=False)

        user.send_password_reset_email()
        reset_link = re.compile(r'http://.*reset/.*/')
        matches = reset_link.findall(mail.outbox[0].body)
        self.assertEqual(len(matches), 1)
        reset_url = matches[0]
        response = self.client.get(reset_url)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b'<form', response.content)

        # form submit
        response = self.client.post(reset_url, {
            'new_password1': 'imnew', 'new_password2': 'imnew'})
        self.assertEqual(response.status_code, 302)
        response = self.client.get(response['Location'])

        self.assertEqual(response.status_code, 200)

        login_ok = self.client.login(
            identifier=user.identifier, password='imnew')
        self.assertTrue(login_ok)



from django.contrib.auth import authenticate
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.authentication import BasicAuthentication

from munch.apps.users.models import APIApplication


class SecretBasicAuthentication(BasicAuthentication):
    """ Uses HTTP basic scheme to login with API key

    API key is bound to the MunchUser model.
    """
    def authenticate_credentials(self, username, secret):
        """
        Authenticate the userid and secret against username and secret.

        Tightly bound to .backends.APIKeyMunchUserBackend
        """
        if username != "api":
            raise AuthenticationFailed('Invalid username')

        user = authenticate(secret=secret)
        if user and user.is_active:
            return (user, None)
        api_app = APIApplication.objects.filter(
            secret=secret, author__is_active=True).first()
        if api_app and api_app.author.is_active:
            return (api_app.author, None)

        raise AuthenticationFailed('Invalid API key')



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-22 16:48
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import munch.core.utils.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('core', '0002_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='OptOut',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(db_index=True, max_length=150, unique=True, verbose_name='identifier')),
                ('address', models.EmailField(max_length=254, verbose_name='mail')),
                ('creation_date', models.DateTimeField(default=django.utils.timezone.now)),
                ('origin', models.CharField(choices=[('mail', 'Email'), ('web', 'Web link'), ('feedback-loop', 'Detected as spam'), ('abuse', 'By abuse report'), ('bounce', 'Too much delivering errors')], max_length=20, verbose_name='origine')),
                ('author', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='author')),
                ('category', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Category', verbose_name='category')),
            ],
            options={
                'verbose_name': 'optout',
                'abstract': False,
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations'),
                'verbose_name_plural': 'optouts',
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
    ]



from django.db import models
from django.utils.translation import ugettext_lazy as _


class DomainCheckField(models.CharField):
    """ To reflects the status of a configuration check on a domain. """
    OK = 'ok'
    NOT_CONFIGURED = 'ko'
    BADLY_CONFIGURED = 'bad'
    PENDING = 'pending'
    UNKNOWN = 'unknown'

    choices = (
        (OK, _('Configured')),
        (NOT_CONFIGURED, _('Not configured')),
        (BADLY_CONFIGURED, _('Badly configured')),
        (PENDING, _('Checking')),
        (UNKNOWN, _('Unknown')))

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 10
        kwargs['choices'] = self.choices
        kwargs['default'] = self.UNKNOWN
        super().__init__(*args, **kwargs)



import uuid
import base64
import time
from math import floor
from socket import getfqdn

from django.conf import settings

from munch.apps.domains.fields import DomainCheckField

from .parsers import extract_domain  # noqa
from .parsers import parse_email_date  # noqa
from .parsers import UniqueEmailAddressParser  # noqa


def mk_base64_uuid(prefix=''):
    """
    :return: a base64url encoded uuid
    """
    r_uuid = base64.urlsafe_b64encode(uuid.uuid4().bytes)
    return '{}{}'.format(prefix, r_uuid.decode('ascii').strip('='))


def mk_msgid():
    """
    :return: a unique Message-ID header content
    """
    msgid_domain = settings.MSGID_DOMAIN or getfqdn()
    return '<{0}.{1:.0f}@{2}>'.format(
        uuid.uuid4().hex, floor(time.time()), msgid_domain)


def get_app_url(domain=None, organization=None):
    if domain and domain.app_domain and \
            domain.app_domain_status == DomainCheckField.OK:
        return 'http://{}'.format(domain.app_domain)
    elif organization and organization.settings.nickname:
        scheme, domain = settings.APPLICATION_URL.strip('/').split('://')
        return '{}://{}.{}'.format(
            scheme, organization.settings.nickname, domain)
    elif (organization and organization.parent and
          organization.parent.settings.nickname):
        scheme, domain = settings.APPLICATION_URL.strip('/').split('://')
        return '{}://{}.{}'.format(
            scheme, organization.settings.nickname, domain)
    else:
        return settings.APPLICATION_URL.strip('/')



from django.conf import settings
from django.utils.crypto import constant_time_compare
from django.utils.http import base36_to_int

from django.contrib.auth.tokens import PasswordResetTokenGenerator


class MunchUserTokenGenerator(PasswordResetTokenGenerator):
    """
    Based on PasswordResetTokenGenerator (django/contrib/auth/tokens.py)
    Just overwritten to accept arbitrary timeout duration.
    """
    key_salt = settings.SECRET_KEY

    def __init__(self, timeout=settings.PASSWORD_RESET_TIMEOUT_DAYS):
        self.timeout = timeout

    def check_token(self, user, token):
        # Parse the token
        try:
            ts_b36, hash = token.split("-")
        except ValueError:
            return False

        try:
            ts = base36_to_int(ts_b36)
        except ValueError:
            return False

        # Check that the timestamp/uid has not been tampered with
        if not constant_time_compare(
                self._make_token_with_timestamp(user, ts), token):
            return False

        # Check the timestamp is within limit
        if (self._num_days(self._today()) - ts) > \
                self.timeout:
            return False

        return True


token_generator = MunchUserTokenGenerator()



import lxml.html.clean
from lxml import etree
from lxml.html import defs
from lxml.html import xhtml_to_html
from django.template import Context
from django.template import Template
from premailer import Premailer


def css_inline_html(html, **kwargs):
    """ Contentfilter to inline CSS into HTML """
    p = Premailer(
        html=html,
        preserve_internal_links=True,
        exclude_pseudoclasses=True,
        keep_style_tags=False,
        include_star_selectors=True,
        remove_classes=False,
        strip_important=False,
        method='html',
        base_path=None,
        disable_basic_attributes=[],
        disable_validation=False,
        disable_leftover_css=True,
        align_floating_images=True,
        remove_unset_properties=True,
        capitalize_float_margin=True)
    return p.transform(pretty_print=True)


class ConditionalFriendlyCleaner(lxml.html.clean.Cleaner):
    def kill_conditional_comments(self, doc):
        # Do nothing, as we want to keep conditionals
        pass

    def kill_nonconditional_comments(self, doc):
        """ Removes all comments but IE conditionals
        """
        for el in doc.iter(etree.Comment):
            if lxml.html.clean._conditional_comment_re.match(el.text) is None:
                el.drop_tree()

    def __call__(self, doc):
        # Needed to overwrite just because of weird behaviour here:
        # https://github.com/lxml/lxml/blob/11e7f9a7d98e96ad31a1d5991b57d47472c1d143/src/lxml/html/clean.py#L308-L311
        # Let's hope it will becone unnecessary,
        # see https://github.com/lxml/lxml/pull/177

        if hasattr(doc, 'getroot'):
            # ElementTree instance, instead of an element
            doc = doc.getroot()
        # convert XHTML to HTML
        xhtml_to_html(doc)
        # Normalize a case that IE treats <image> like <img>, and that
        # can confuse either this step or later steps.
        for el in doc.iter('image'):
            el.tag = 'img'

        # BEGIN MUNCH EDIT
        # if not self.comments:
        #     # Of course, if we were going to kill comments anyway, we don't
        #     # need to worry about this
        #     self.kill_conditional_comments(doc)
        # END MUNCH EDIT

        kill_tags = set(self.kill_tags or ())
        remove_tags = set(self.remove_tags or ())
        allow_tags = set(self.allow_tags or ())

        if self.scripts:
            kill_tags.add('script')
        if self.safe_attrs_only:
            safe_attrs = set(self.safe_attrs)
            for el in doc.iter(etree.Element):
                attrib = el.attrib
                for aname in attrib.keys():
                    if aname not in safe_attrs:
                        del attrib[aname]
        if self.javascript:
            if not (self.safe_attrs_only and
                    self.safe_attrs == defs.safe_attrs):
                # safe_attrs handles events attributes itself
                for el in doc.iter(etree.Element):
                    attrib = el.attrib
                    for aname in attrib.keys():
                        if aname.startswith('on'):
                            del attrib[aname]
            doc.rewrite_links(self._remove_javascript_link,
                              resolve_base_href=False)
            if not self.style:
                # If we're deleting style then we don't have to remove JS links
                # from styles, otherwise...
                for el in lxml.html.clean._find_styled_elements(doc):
                    old = el.get('style')
                    new = lxml.html.clean._css_javascript_re.sub('', old)
                    new = lxml.html.clean._css_import_re.sub('', new)
                    if self._has_sneaky_javascript(new):
                        # Something tricky is going on...
                        del el.attrib['style']
                    elif new != old:
                        el.set('style', new)
                for el in list(doc.iter('style')):
                    if el.get('type', '').lower().strip() == 'text/javascript':
                        el.drop_tree()
                        continue
                    old = el.text or ''
                    new = lxml.html.clean._css_javascript_re.sub('', old)
                    # The imported CSS can do anything; we just can't allow:
                    new = lxml.html.clean._css_import_re.sub('', old)
                    if self._has_sneaky_javascript(new):
                        # Something tricky is going on...
                        el.text = '/* deleted */'
                    elif new != old:
                        el.text = new
        # START MUNCH MODIFICATION
        if self.comments:
            # kill_tags.add(etree.Comment)
            self.kill_nonconditional_comments(doc)
        # END MUNCH MODIFICATION
        if self.processing_instructions:
            kill_tags.add(etree.ProcessingInstruction)
        if self.style:
            kill_tags.add('style')
            etree.strip_attributes(doc, 'style')
        if self.links:
            kill_tags.add('link')
        elif self.style or self.javascript:
            # We must get rid of included stylesheets if Javascript is not
            # allowed, as you can put Javascript in them
            for el in list(doc.iter('link')):
                if 'stylesheet' in el.get('rel', '').lower():
                    # Note this kills alternate stylesheets as well
                    if not self.allow_element(el):
                        el.drop_tree()
        if self.meta:
            kill_tags.add('meta')
        if self.page_structure:
            remove_tags.update(('head', 'html', 'title'))
        if self.embedded:
            # FIXME: is <layer> really embedded?
            # We should get rid of any <param> tags not inside <applet>;
            # These are not really valid anyway.
            for el in list(doc.iter('param')):
                parent = el.getparent()
                while parent is not None and parent.tag not in (
                        'applet', 'object'):
                    parent = parent.getparent()
                if parent is None:
                    el.drop_tree()
            kill_tags.update(('applet',))
            # The alternate contents that are in an iframe are a good fallback:
            remove_tags.update(('iframe', 'embed', 'layer', 'object', 'param'))
        if self.frames:
            # FIXME: ideally we should look at the frame links, but
            # generally frames don't mix properly with an HTML
            # fragment anyway.
            kill_tags.update(defs.frame_tags)
        if self.forms:
            remove_tags.add('form')
            kill_tags.update(('button', 'input', 'select', 'textarea'))
        if self.annoying_tags:
            remove_tags.update(('blink', 'marquee'))

        _remove = []
        _kill = []
        for el in doc.iter():
            if el.tag in kill_tags:
                if self.allow_element(el):
                    continue
                _kill.append(el)
            elif el.tag in remove_tags:
                if self.allow_element(el):
                    continue
                _remove.append(el)

        if _remove and _remove[0] == doc:
            # We have to drop the parent-most tag, which we can't
            # do.  Instead we'll rewrite it:
            el = _remove.pop(0)
            el.tag = 'div'
            el.attrib.clear()
        elif _kill and _kill[0] == doc:
            # We have to drop the parent-most element, which we can't
            # do.  Instead we'll clear it:
            el = _kill.pop(0)
            if el.tag != 'html':
                el.tag = 'div'
            el.clear()

        _kill.reverse()  # start with innermost tags
        for el in _kill:
            el.drop_tree()
        for el in _remove:
            el.drop_tag()

        if self.remove_unknown_tags:
            if allow_tags:
                raise ValueError((
                    "It does not make sense to pass in both "
                    "allow_tags and remove_unknown_tags"))
            allow_tags = set(defs.tags)
        if allow_tags:
            bad = []
            for el in doc.iter():
                if el.tag not in allow_tags:
                    bad.append(el)
            if bad:
                if bad[0] is doc:
                    el = bad.pop(0)
                    el.tag = 'div'
                    el.attrib.clear()
                for el in bad:
                    el.drop_tag()

        if self.add_nofollow:
            for el in lxml.html.clean._find_external_links(doc):
                if not self.allow_follow(el):
                    rel = el.get('rel')
                    if rel:
                        if (
                                'nofollow' in rel and ' nofollow ' in (
                                    ' %s ' % rel)):
                            continue
                        rel = '%s nofollow' % rel
                    else:
                        rel = 'nofollow'
                    el.set('rel', rel)


# Waiting for https://github.com/lxml/lxml/pull/178 integration for comment
# tbody removal is linked to
# https://blog.mailchimp.com/spam-filters-dont-like-microsoft/

_allowed_tags = (
    defs.tags | frozenset([etree.Comment])) \
    - frozenset(['tbody', 'thead', 'tfoot'])
cleaner = ConditionalFriendlyCleaner(
    scripts=True,
    javascript=True,
    embedded=True,
    safe_attrs_only=False,
    links=True,
    page_structure=False,  # avoid wrapping everything in a div...
    comments=True,  # Non-conditionals only
    allow_tags=_allowed_tags,
    remove_unknown_tags=False)


def clean_html(html, **kwargs):
    """ Contentfilter to remove js & others from HTML """
    return cleaner.clean_html(html)


def apply_mailmerge(html, mail_properties={}, **kwargs):
    tpl = Template(html)
    return tpl.render(Context(mail_properties))



from django import forms
from django.utils.translation import ugettext_lazy as _

from munch.apps.campaigns.models import Mail

from .models import AbuseNotification


class AbuseNotificationForm(forms.ModelForm):
    mail = forms.CharField(max_length=50, widget=forms.HiddenInput())
    contact_name = forms.CharField(
        required=False,
        widget=forms.TextInput(attrs={'placeholder': _('(Optional)')}))
    contact_email = forms.CharField(
        required=False,
        widget=forms.TextInput(attrs={'placeholder': _('(Optional)')}))
    comments = forms.CharField(
        widget=forms.Textarea(attrs={
            'placeholder': _('Comments, questions, reason for report'),
            'rows': 4
        }))

    class Meta:
        model = AbuseNotification
        exclude = ['date']

    def clean_mail(self):
        return Mail.objects.get(identifier=self.cleaned_data['mail'])



from ...models import Category
from ...utils.filters import DRFFilterSet


class CategoryFilter(DRFFilterSet):
    class Meta:
        model = Category
        fields = ['name']
        order_by = True



import django_filters

from munch.core.utils.filters import DRFFilterSet

from ...models import Mail
from ...models import Message
from ...models import MailStatus


class MessageFilter(DRFFilterSet):
    status = django_filters.MultipleChoiceFilter(
        choices=Message.STATUS_CHOICES)

    class Meta:
        model = Message
        fields = [
            'creation_date', 'status', 'category', 'sender_email',
            'track_open', 'track_clicks', 'external_optout', 'is_spam']
        # default order by -creation_date
        order_by = ['-' + i for i in fields] + [i for i in fields]


class MailFilter(DRFFilterSet):
    last_status = django_filters.CharFilter(name='curstatus')
    opened = django_filters.MethodFilter(action='_opened')
    clicked = django_filters.MethodFilter(action='_clicked')
    to = django_filters.CharFilter(name='recipient')
    delivery_status = django_filters.MultipleChoiceFilter(
        name='curstatus', choices=MailStatus.STATUS_CHOICES)

    class Meta:
        model = Mail
        # TODO: To must be recipient in v2
        fields = {
            'to': ['exact'],
            'last_status': ['exact'],
            'delivery_status': ['exact'],
            'source_ref': ['exact', 'startswith'],
            'source_type': ['exact']}
        order_by = True

    def _opened(self, queryset, value):
        # Watch out, putting the same name for method and filter field
        # overwrites the field
        return queryset.opened()

    def _clicked(self, queryset, value):
        return queryset.clicked()


class MailStatusFilter(DRFFilterSet):
    address = django_filters.CharFilter(name='mail__recipient')

    class Meta:
        model = MailStatus
        fields = ['address', 'creation_date', 'status']
        order_by = True



from django.db.models import Prefetch
from rest_framework import viewsets
from rest_framework.response import Response
from rest_framework.decorators import detail_route

from munch.core.utils.views import filtered
from munch.core.utils.views import paginated
from munch.core.utils.views import MunchModelViewSetMixin
from munch.core.utils.views import OrganizationOwnedViewSetMixin
from munch.apps.optouts.models import OptOut
from munch.apps.optouts.api.v1.filters import OptOutFilter
from munch.apps.optouts.api.v1.serializers import OptOutSerializer

from ...models import Mail
from ...models import MailBatch
from ...models import MailStatus
from .filters import MailFilter
from .filters import MailStatusFilter
from .serializers import MailSerializer
from .serializers import MailBatchSerializer
from .serializers import MailStatusSerializer


class MailViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ReadOnlyModelViewSet):
    model = Mail
    serializer_class = MailSerializer
    filter_class = MailFilter

    def get_queryset(self):
        status_qs = MailStatus.objects.order_by(
            'mail', '-creation_date').distinct('mail')
        return super().get_queryset().prefetch_related(
            Prefetch(
                'statuses', queryset=status_qs, to_attr='last_status_cached'))

    @detail_route(methods=['get'])
    def statuses(self, request, pk):
        serializer = MailStatusSerializer(
            self.get_object().statuses.all(),
            many=True, context={'request': request})
        return Response(serializer.data)


class MailBatchViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ReadOnlyModelViewSet):
    model = MailBatch
    serializer_class = MailBatchSerializer

    @detail_route(methods=['get'])
    @paginated(MailSerializer)
    @filtered(MailFilter)
    def mails(self, request, pk):
        return self.get_object().mails.all()

    @detail_route(methods=['get'])
    def stats(self, request, pk, format=None):
        batch = self.get_object()
        return Response(batch.mk_stats())

    @detail_route(methods=['get'])
    @paginated(MailStatusSerializer)
    @filtered(MailStatusFilter)
    def bounces(self, request, pk):
        batch = self.get_object()
        return MailStatus.objects.filter(
            mail__in=batch.mails.all().only('id'),
            status__in=(MailStatus.BOUNCED, MailStatus.DROPPED))

    @detail_route(methods=['get'])
    @paginated(OptOutSerializer)
    @filtered(OptOutFilter)
    def opt_outs(self, request, pk, format=None):
        identifiers = Mail.objects.filter(
            batch=self.get_object()).values_list('identifier', flat=True)
        return OptOut.objects.filter(identifier__in=identifiers)



from rest_framework import permissions

from munch.core.utils.permissions import MunchResourcePermission

from ...models import Message


class AttachmentPermission(permissions.BasePermission):
    """ Permission is granted to attach files only to specified users, others
    can view their attachment (if any), no more.
    """
    def has_permission(self, request, view):
        if request.method in ['OPTIONS', 'HEAD']:
            return True
        elif request.user.is_authenticated():
            if request.method == 'GET':
                return True
            if request.user.is_staff or \
                    request.user.organization.can_attach_files:
                return True
        return False


class PreviewMailPermission(MunchResourcePermission):
    app_name = 'campaigns'
    model_name = 'message'

    def has_object_permission(self, request, view, obj):
        # checks for message_add
        allowed = super().has_object_permission(request, view, obj)

        if allowed:
            return self.can_user_do(request.user, obj, 'previewsend')
        else:
            return False


class MailPermission(MunchResourcePermission):
    def has_object_permission(self, request, view, obj):
        if request.method == 'DELETE':
            if obj.message.status in [Message.SENDING, Message.SENT]:
                return False
        return super().has_object_permission(request, view, obj)



from django.conf import settings
from django.conf.urls import url
from django.urls import reverse_lazy

from .tokens import MunchUserTokenGenerator
from .forms import InvitationForm
from . import views

urlpatterns = [
    # Invitation
    url((
        r'^invitation/confirm/(?P<uidb64>[0-9A-Za-z_\-]+)/'
        r'(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,20})/$'),
        views.password_reset_confirm,
        {
            'set_password_form': InvitationForm,
            'post_reset_redirect': reverse_lazy('invitation_complete'),
            'token_generator': MunchUserTokenGenerator(
                settings.INVITATION_TIMEOUT_DAYS),
            'template_name': 'users/invitation_confirm.html', 'initial': True},
        name='invitation_confirm'),
    url(
        r'^invitation/complete/$',
        views.password_reset_complete,
        {'template_name': 'users/invitation_complete.html'},
        name='invitation_complete'),
    # Password Reset
    url((
        r'^password/reset/(?P<uidb64>[0-9A-Za-z_\-]+)/'
        r'(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,20})/$'),
        views.password_reset_confirm,
        {
            'post_reset_redirect': reverse_lazy('password_reset_complete'),
            'token_generator': MunchUserTokenGenerator(),
            'template_name': 'users/password_reset_confirm.html'},
        name='password_reset_confirm'),
    url(
        r'^password/reset/complete/$',
        views.password_reset_complete,
        {'template_name': 'users/password_reset_complete.html'},
        name='password_reset_complete'),
]



import os
import copy
import base64
import shutil
import inspect
from functools import wraps


class HttpBasicTestCaseMixin:
    """ Just a way to build authentication headers

    Example usage :
      self.client.get('/foo',
                      **self.http_basic_headers('user', 'password'))
    """
    def http_basic_headers(self, username, password):
        key = base64.b64encode(
            '{}:{}'.format(username, password).encode('utf-8')).decode()
        return {'HTTP_AUTHORIZATION': 'Basic {}'.format(key)}


class temporary_settings:
    def __enter__(self):
        self.enable()

    def __exit__(self, exc_type, exc_value, traceback):
        self.disable()

    def __call__(self, test_func):
        from django.test import SimpleTestCase
        if isinstance(test_func, type):
            if not issubclass(test_func, SimpleTestCase):
                raise Exception(
                    "Only subclasses of Django SimpleTestCase "
                    "can be decorated with temporary_settings")
            return test_func
        else:
            @wraps(test_func)
            def inner(*args, **kwargs):
                with self:
                    return test_func(*args, **kwargs)
        return inner

    def enable(self):
        from django.conf import settings
        self.wrapped = copy.deepcopy(settings)

    def disable(self):
        from django.conf import settings
        settings._wrapped = self.wrapped



import logging

from celery import task

log = logging.getLogger(__name__)


@task
def purge_raw_mail():
    from .models import RawMail
    kwargs = {}
    for related_object in RawMail._meta.related_objects:
        accessor_name = related_object.field.related_query_name()
        kwargs.update({'{}__isnull'.format(accessor_name): True})
        log.info('Scanning all orphans RawMail.{}...'.format(accessor_name))

    count, _ = RawMail.objects.filter(**kwargs).only('id').delete()
    log.info('RawMail deleted: {} ({})'.format(count, kwargs))



import re
import uuid
import time
from math import floor
from email.generator import Generator
from email.utils import formatdate

from jinja2 import Template
from jinja2 import Environment
from jinja2 import PackageLoader

from .parsers import parse_email_date
from ..models import AbstractMailStatus


class InvalidDSNError(ValueError):
    def __init__(self, msg):
        self.msg = 'Invalid DSN: {}'.format(msg)


class DSNParser:
    """ Delivery status notification

    As described by RFC3464

    Note that only individual DSN (for one recipient) is handled.
    """
    ACTION_DELAYED = 'delayed'
    ACTION_EXPANDED = 'expanded'
    ACTION_RELAYED = 'relayed'
    ACTION_FAILED = 'failed'
    ACTION_DELIVERED = 'delivered'

    r_diagnostic_code = re.compile(
        r'smtp;\s*(?P<smtp_code>\d{3})[ -].*')

    # Map linking mail statuses and DSN esmtp code first digit
    STATUS_MAP = {
        '2': AbstractMailStatus.DELIVERED,
        '5': AbstractMailStatus.BOUNCED,
        '4': AbstractMailStatus.DROPPED
    }

    def __init__(self, headers=None):
        """ Represents/parses the multipart/report

        :type headers: dict
        :param headers: a dict containing DSN headers
        """
        self.headers = headers
        if headers:
            self.parse_headers(headers)

    def parse_headers(self, headers):
        if not self.headers:
            self.headers = headers
        # Check for mandatory DSN report headers
        try:
            self.final_recipient = headers['Final-Recipient']
        except KeyError:
            raise InvalidDSNError('missing Final-Recipient header')
        try:
            self.esmtp_status = headers['Status']
        except KeyError:
            raise InvalidDSNError('missing Status header')
        try:
            self.action = headers.get('Action', '').lower()
        except KeyError:
            raise InvalidDSNError('missing Action header')

        self.msg = headers.get('Diagnostic-Code', '').replace('\n', ' ')
        m = self.r_diagnostic_code.match(self.msg)
        if m:
            self.smtp_status = m.group('smtp_code')
        else:
            self.smtp_status = None

        try:
            self.original_recipient = headers['Original-Recipient']
        except KeyError:
            self.original_recipient = None

        try:
            date_str = headers['Arrival-Date']
        except KeyError:
            self.date = None
        else:
            self.date = parse_email_date(date_str)

    def get_next_mailstatus(self):
        # there are two cases with 4.x.x errors : retry and given up
        if self.action == self.ACTION_DELAYED:
            return AbstractMailStatus.DELAYED

        else:
            try:
                esmtp_first_digit = self.esmtp_status.split('.')[0]
                return self.STATUS_MAP[esmtp_first_digit]
            except KeyError:
                raise ValueError('Invalid DSN: unknown code : {}'.format(
                    self.esmtp_status))

    def as_string(self):
        raw_dsn = ""
        for key, value in self.headers.items():
            raw_dsn += "{}: {}\n".format(key, value)
        return raw_dsn


DEFAULT_SUBJECTS = {
    '2': 'Successful Mail Delivery Report',
    '5': 'Undelivered Mail Returned to Sender',
    '4': 'Undelivered Mail Returned to Sender',
}
DEFAULT_REPORT_DESCRIPTION = {
    '2': 'Delivered Message',
    '5': 'Undelivered Message',
    '4': 'Undelivered Message'
}
ACTIONS_STATUS_MAP = {
    '2': 'delivered',
    '4': 'failed',
    '5': 'failed'
}


class DSN:
    """ Delivery status notification

    As described by RFC3464

    Note that only individual DSN (for one recipient) is handled.
    """

    def __init__(self, dsn_from, dsn_to, subject=None):

        self.context = {
            'boundary': Generator._make_boundary(),
            'from': dsn_from,
            'to': dsn_to,
            'subject': subject,
        }

    def generate(
            self, original_headers, original_body, reply_code, reply_message,
            reporting_mta, remote_mta, date, report_headers={}, template=None):

        # Main part
        timestamp = floor(time.time())
        self.context.update({
            'date': formatdate(localtime=True),
            'identifier': '<{0}.{1:.0f}@{2}>'.format(
                uuid.uuid4().hex, floor(timestamp), reporting_mta),
        })
        if not self.context['subject']:
            self.context.update({
                'subject': DEFAULT_SUBJECTS.get(reply_code[0], None)
            })
        # Report & body
        self.context.update({
            'arrival_date': date,
            'reporting_mta': reporting_mta,
            'remote_mta': remote_mta,
            'report_headers': report_headers,
            'reply_code': reply_code,
            'reply_status': reply_message.split(' ')[0],
            'reply_message': reply_message,
            'action': ACTIONS_STATUS_MAP.get(reply_code[0], 'failed'),
            'original_recipient': original_headers['To'],
        })
        # Original message
        self.context.update({
            'description': DEFAULT_REPORT_DESCRIPTION.get(reply_code[0]),
            'original_headers': original_headers,
            'original_body': original_body.decode('utf-8'),
        })

        environment = Environment(
            loader=PackageLoader('munch.core.mail.utils', 'templates'))
        dsn_template = environment.get_template('dsn.jinja2')

        if template is None:
            dsn_body_template = environment.get_template('dsn.body.jinja2')
        else:
            dsn_body_template = Template(template)

        self.context.update(
            {'dsn_body': dsn_body_template.render(self.context)})

        message = dsn_template.render(self.context)
        return message.encode('utf-8')



from django.db import models
from django.utils.translation import ugettext_lazy as _

from munch.core.utils.models import AbstractOwnedModel

from .signals import pre_validation
from .signals import post_validation


class Category(AbstractOwnedModel):
    name = models.CharField(_('name'), max_length=100)
    author = models.ForeignKey('users.MunchUser', verbose_name=_('author'))

    owner_path = 'author__organization'
    author_path = 'author'

    class Meta(AbstractOwnedModel.Meta):
        verbose_name = _('category')
        verbose_name_plural = _('categories')

    def __str__(self):
        return self.name

    def mk_stats(self):
        from munch.apps.optouts.models import OptOut
        from munch.apps.campaigns.models import Mail as CampaignMail

        campaign_mail_qs = CampaignMail.objects.filter(message__category=self)
        optout_qs = OptOut.objects.filter(
            identifier__in=campaign_mail_qs.values_list(
                'identifier', flat=True))

        return {
            'count': campaign_mail_qs.info_counts(),
            'last_status': campaign_mail_qs.last_status_counts(),
            'timing': {
                'delivery_median': campaign_mail_qs.median(
                    'delivery_duration').seconds
            },
            'optout': optout_qs.count_by_origin(with_total=True)
        }


class ValidationSignalsModel:
    def clean(self, *args, **kwargs):
        created = True
        if self.pk:
            created = False

        pre_validation.send(
            sender=self.__class__, instance=self, created=created)
        super().clean(*args, **kwargs)
        post_validation.send(
            sender=self.__class__, instance=self, created=created)



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import PermissionDenied


class SecretMunchUserBackend(ModelBackend):
    def authenticate(self, secret=None, **kwargs):
        """ Tries to authenticate against secret.

        Fails silently if there is no secret provided or
        Reject the authentication if there is an secret
        provided but its wrong.
        """
        UserModel = get_user_model()
        if secret:
            try:
                user = UserModel._default_manager.get(secret=secret)
                return user
            except UserModel.DoesNotExist:
                raise PermissionDenied('Invalid API credentials')



import re

import dns
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext as _

# Regexp matching valid SPF record
r_spf = re.compile(r'"v=spf1 .*[\?\-~]all"')


class SPFValidator:
    """ Just a tool to validate SPF records
    """

    def __init__(self, spf_include):
        """
        :param spf_include: a domain that should be spf-included
                            for the record to be good.
        """
        self.spf_include = spf_include

    def _fetch_spf_record(self, domain_name):
        """ Looks for SPF record in all TXT and SPF records

        :return the first valid SPF field value found or None
        """
        for field in ('TXT', 'SPF'):
            try:
                res = dns.resolver.query(domain_name, field)
            except dns.exception.DNSException:
                # try next record type
                pass
            else:
                for field in res:
                    field_text = field.to_text()
                    if r_spf.match(field_text):
                        return field_text

    def validate(self, domain_name):
        include = 'include:{}'.format(self.spf_include)
        field_text = self._fetch_spf_record(domain_name)
        if field_text is None:
            raise ValidationError(
                _("SPF not configured for {} domain").format(domain_name))
        if include in field_text:
            return domain_name
        else:
            raise ValidationError(_(
                "SPF badly configured for {} domain").format(domain_name))



import re
import unittest
from datetime import date
from datetime import timedelta

from django.core import mail
from django.urls import reverse
from django.conf import settings
from django.test import TestCase
from django.utils import timezone
from django.test.utils import override_settings
from libfaketime import fake_time
from rest_framework.test import APIClient

from .. import tasks
from ..models import Contact
from ..models import ContactList
from ..models import ContactQueue
from ..models import CollectedContact
from ..models import ContactListPolicy
from ..models import ContactListPolicyAttribution
from ..models import ContactQueuePolicyAttribution

from munch.apps.users.tests.factories import UserFactory


bounce = {
    'Auto-Submitted': 'auto-replied',
    'Content-Type': 'multipart/report; report-type=delivery-status;\n\t'
    'boundary="A10868308A6BB.1438776632/mail.example.com"',
    'Date': 'Wed,  5 Aug 2015 14:10:32 +0200 (CEST)',
    'Delivered-To': '{bounce}',
    'From': 'MAILER-DAEMON@example.com (Mail Delivery System)',
    'MIME-Version': '1.0',
    'Message-Id': '<20150805121032.E77C18308A6BC@mail.example.com>',
    'Received': 'by mail.example.com (Postfix)\n\tid E77C18308A6BC;'
    ' Wed,  5 Aug 2015 14:10:32 +0200 (CEST)',
    'Return-Path': '<MAILER-DAEMON>',
    'Subject': 'Undelivered Mail Returned to Sender',
    'To': '{bounce}'}


URL_REGEX = re.compile('https?://[^\s]+')


@override_settings(
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    CELERY_ALWAYS_EAGER=True,
    BROKER_BACKEND='memory')
class ContactQueuesTestCase(TestCase):
    fixtures = ['generic_policies']

    def setUp(self):
        self.user = UserFactory(groups=['managers'])
        # Remove subscription email
        mail.outbox = []
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

    def test_subscription_without_policies(self):
        q = ContactQueue(author=self.user)
        q.save()

        contact = CollectedContact(
            contact_queue=q, address='nope@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        self.assertEqual(CollectedContact.OK, contact.status)
        self.assertEqual(0, len(mail.outbox))

    def test_single_policy_subscription(self):
        q = ContactQueue(author=self.user)
        q.save()

        qa = ContactQueuePolicyAttribution(
            contact_queue=q,
            policy=ContactListPolicy.objects.get(name='DoubleOptIn'))
        qa.save()

        contact = CollectedContact(
            contact_queue=q, address='nope@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        self.assertEqual(CollectedContact.PENDING, contact.status)
        self.assertEqual(1, len(mail.outbox))

    @unittest.skip('Disabled temporarily to allow deploy. FIXME !!!')
    def test_bounce_backend(self):
        q = ContactQueue(author=self.user)
        q.save()

        qa = ContactQueuePolicyAttribution(
            contact_queue=q,
            policy=ContactListPolicy.objects.get(name='BounceCheck'))
        qa.save()

        contact = CollectedContact(
            contact_queue=q, address='nope@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        self.assertEqual(1, len(mail.outbox))

        bounce_address = mail.outbox.pop().message().get('Return-Path')

        # sending bounce task
        tasks.handle_bounce(
            {k: v.format(bounce=bounce_address) for k, v in bounce.items()})

        contact.refresh_from_db()
        self.assertEqual(CollectedContact.BOUNCED, contact.status)

        contact = CollectedContact(
            contact_queue=q, address='truemail@example.com',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        with fake_time(timezone.now() + timedelta(hours=5)):
            tasks.handle_bounce_expirations()

        contact.refresh_from_db()
        self.assertEqual(CollectedContact.OK, contact.status)

        contact = CollectedContact(
            contact_queue=q, address='truemail2@example.com',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        with fake_time(timezone.now() + timedelta(minutes=5)):
            tasks.handle_bounce_expirations()

        contact.refresh_from_db()
        self.assertEqual(
            CollectedContact.PENDING, contact.status,
            'Bounce-check hasn’t expired yet, so the contact should not be OK')

    def test_double_opt_in_backend(self):
        q = ContactQueue(author=self.user)
        q.save()

        qa = ContactQueuePolicyAttribution(
            contact_queue=q,
            policy=ContactListPolicy.objects.get(name='DoubleOptIn'))
        qa.save()

        contact = CollectedContact(
            contact_queue=q, address='nope@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        self.assertEqual(1, len(mail.outbox))
        confirmation_msg = mail.outbox.pop().body
        url = URL_REGEX.search(confirmation_msg).group().replace(
            settings.APPLICATION_URL, '')
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)

        contact.refresh_from_db()
        self.assertEqual(CollectedContact.OK, contact.status)

        contact = CollectedContact(
            contact_queue=q, address='nope2@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        with fake_time(timezone.now() + timedelta(days=8)):
            tasks.handle_opt_ins_expirations()

        contact.refresh_from_db()
        self.assertEqual(CollectedContact.EXPIRED, contact.status)

        contact = CollectedContact(
            contact_queue=q, address='nope3@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        with fake_time(timezone.now() + timedelta(days=2)):
            tasks.handle_opt_ins_expirations()

        contact.refresh_from_db()
        self.assertEqual(CollectedContact.PENDING, contact.status)

    @unittest.skip('Disabled temporarily to allow deploy. FIXME !!!')
    def test_multiple_backends(self):
        q = ContactQueue(author=self.user)
        q.save()

        qa1 = ContactQueuePolicyAttribution(
            contact_queue=q,
            policy=ContactListPolicy.objects.get(name='DoubleOptIn'))
        qa1.save()

        qa2 = ContactQueuePolicyAttribution(
            contact_queue=q,
            policy=ContactListPolicy.objects.get(name='BounceCheck'))
        qa2.save()

        contact = CollectedContact(
            contact_queue=q, address='nope1@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        self.assertEqual(CollectedContact.PENDING, contact.status)
        self.assertEqual(
            1, len(mail.outbox),
            'Using DoubleOptIn and BounceCheck should only send 1 mail')

        bounce_address = mail.outbox.pop().message().get('Return-Path')

        # sending bounce task
        tasks.handle_bounce(
            {k: v.format(bounce=bounce_address) for k, v in bounce.items()})

        contact.refresh_from_db()
        self.assertEqual(CollectedContact.BOUNCED, contact.status)

        contact = CollectedContact(
            contact_queue=q, address='nope2@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        confirmation_msg = mail.outbox.pop().body
        url = URL_REGEX.search(confirmation_msg).group().replace(
            settings.APPLICATION_URL, '')
        self.client.get(url)

        contact.refresh_from_db()
        self.assertEqual(
            CollectedContact.OK, contact.status,
            'Contact should be validated if confirmation link is '
            'clicked, even if the bounce hasn’t expired')

        contact = CollectedContact(
            contact_queue=q, address='nope3@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        with fake_time(timezone.now() + timedelta(hours=5)):
            tasks.handle_bounce_expirations()

        contact.refresh_from_db()
        self.assertEqual(
            CollectedContact.PENDING, contact.status,
            'Bounce expiration should not set contact status to '
            '“ok” if DoubleOptIn is set')

    @unittest.skip('Disabled temporarily to allow deploy. FIXME !!!')
    def test_failed_expiration(self):
        q = ContactQueue(author=self.user)
        q.save()

        qa = ContactQueuePolicyAttribution(
            contact_queue=q,
            policy=ContactListPolicy.objects.get(name='DoubleOptIn'))
        qa.save()

        contact = CollectedContact(
            contact_queue=q, address='nope@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        with fake_time(timezone.now() + timedelta(days=8)):
            tasks.handle_opt_ins_expirations()

        with fake_time(timezone.now() + timedelta(days=15)):
            tasks.handle_failed_expirations()

        # Long-time expired contact should be deleted
        self.assertRaises(contact.DoesNotExist, contact.refresh_from_db)

        contact = CollectedContact(
            contact_queue=q, address='nope@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        q = ContactQueue(author=self.user)
        q.save()

        qa = ContactQueuePolicyAttribution(
            contact_queue=q,
            policy=ContactListPolicy.objects.get(name='BounceCheck'))
        qa.save()

        contact = CollectedContact(
            contact_queue=q, address='bounce@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        bounce_address = mail.outbox.pop().message().get('Return-Path')

        # sending bounce task
        tasks.handle_bounce(
            {k: v.format(bounce=bounce_address) for k, v in bounce.items()})

        with fake_time(timezone.now() + timedelta(days=15)):
            tasks.handle_failed_expirations()

        # Long-time bounced contact should be deleted
        self.assertRaises(contact.DoesNotExist, contact.refresh_from_db)

        contact = CollectedContact(
            contact_queue=q, address='ok@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        contact.status = CollectedContact.CONSUMED
        contact.save()

        with fake_time(timezone.now() + timedelta(days=15)):
            tasks.handle_consumed_contacts_expirations()

        # Long-time consumed contact should be deleted
        self.assertRaises(contact.DoesNotExist, contact.refresh_from_db)

    def test_http_subscription(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'name', 'type': 'Char'},
                {'name': 'age', 'type': 'Integer'}])
        q.save()

        response = self.client.post(
            reverse('subscription', kwargs={'uuid': q.uuid}),
            {'address': 'yep@example.org', 'name': 'John Doe', 'age': 42})
        self.assertEqual(response.status_code, 201)

        try:
            contact = q.collected_contacts.get(address='yep@example.org')
        except CollectedContact.DoesNotExist:
            self.fail('The contact should have been saved')

        self.assertEqual(2, len(contact.properties))

        self.client.post(
            reverse('subscription', kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'name': 'John Doe', 'age': 'won’t tell'})

        self.assertRaises(
            CollectedContact.DoesNotExist,
            q.collected_contacts.get,
            address='nope@example.org')

    def test_queue_creation(self):
        types = self.client.get('/v1/contacts/queues/fields/')
        self.assertEqual(200, types.status_code)
        types = types.data
        self.assertTrue(len(types) >= 3)
        self.assertIn('Char', types)
        self.assertIn('Boolean', types)
        self.assertIn('Date', types)

        policies = self.client.get('/v1/contacts/policies/')
        self.assertEqual(200, policies.status_code)
        policies = policies.data['results']
        self.assertTrue(len(policies) >= 2)

        response = self.client.post(
            '/v1/contacts/queues/',
            {'policies': [policies[0]['url'], policies[1]['url']],
             "contact_fields": [
                 {"name": "prop1", "type": "Char", "required": True},
                 {"name": "prop2", "type": "Boolean", "required": False},
                 {"name": "prop3", "type": "Date", "required": False}
            ]}, format='json')

        queue = response.data
        self.assertEqual(201, response.status_code)
        self.assertEqual(queue['contact_fields'], [
            {"name": "prop1", "type": "Char", "required": True},
            {"name": "prop2", "type": "Boolean", "required": False},
            {"name": "prop3", "type": "Date", "required": False}])
        self.assertEqual(
            queue['policies'], [policies[0]['url'], policies[1]['url']])

        # Test that it appears in listing

        list_response = self.client.get('/v1/contacts/queues/')
        self.assertEqual(list_response.status_code, 200)
        self.assertEqual(list_response.data['results'], [queue])

        # Test that it appears in detail view

        detail_response = self.client.get(queue['url'])
        self.assertEqual(detail_response.data, queue)

        # testing contact add

        contact = self.client.post(
            queue['subscription'], {
                'address': 'nope@example.org',
                'prop1': 'test', 'prop2': True, 'prop3': date.today()})

        self.assertEqual(201, contact.status_code)

        try:
            contact = CollectedContact.objects.get(address='nope@example.org')
        except CollectedContact.DoesNotExist:
            self.fail('Contact should have been written')

        self.assertEqual('127.0.0.1', contact.subscription_ip)

        # should trigger an error the second time

        response = self.client.post(
            queue['subscription'], {
                'address': 'nope@example.org',
                'prop1': 'test', 'prop2': True, 'prop3': timezone.now()})

        self.assertEqual(400, response.status_code)

    def test_queue_creation_bad_proptype(self):
        response = self.client.post(
            '/v1/contacts/queues/',
            {'contact_fields': [
                {'type': 'Char', 'name': 'prop1'},
                {'type': 'Foo', 'name': 'prop2'}],
             'policies': []},
            format='json')
        self.assertEqual(response.status_code, 400)
        self.assertIn('Foo', response.content.decode('utf-8'))

    def test_queue_update(self):
        q = ContactQueue(
            author=self.user,
            properties={'name': 'Char', 'age': 'Integer'})
        q.save()
        ContactQueuePolicyAttribution.objects.create(
            contact_queue=q,
            policy=ContactListPolicy.objects.get(name='DoubleOptIn'))
        response = self.client.put(
            '/v1/contacts/queues/{}/'.format(q.pk), {
                'contact_fields': [
                    {'type': 'Char', 'name': 'lala1'},
                ],
                'policies': ['/v1/contacts/policies/1/',
                             '/v1/contacts/policies/2/']
            }, format='json')
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            response.data['policies'],
            [
                'http://testserver/v1/contacts/policies/1/',
                'http://testserver/v1/contacts/policies/2/'])

    def test_queue_consumption(self):
        q = ContactQueue(author=self.user)
        q.save()

        contact1 = CollectedContact(
            contact_queue=q, address='ok@nope', subscription_ip='127.0.0.1')
        contact1.save()
        contact1.apply_policies()

        contact2 = CollectedContact(
            contact_queue=q, address='pending@nope',
            subscription_ip='127.0.0.1')
        contact2.save()

        consumed = self.client.post(
            '/v1/contacts/queues/{}/consume/'.format(q.pk))

        self.assertEqual(200, consumed.status_code)
        self.assertEqual(1, len(consumed.data['contacts']))
        self.assertEqual('ok@nope', consumed.data['contacts'][0]['address'])

        consumed = self.client.post(
            '/v1/contacts/queues/{}/consume/'.format(q.pk))

        self.assertEqual(0, len(consumed.data['contacts']))

        consumed = self.client.post(
            '/v1/contacts/queues/{}/consume/?status=pending'.format(q.pk))

        self.assertEqual(1, len(consumed.data['contacts']))

        contact2.status = CollectedContact.OK
        contact2.save()

        consumed = self.client.post(
            '/v1/contacts/queues/{}/consume/'.format(q.pk))

        self.assertEqual(1, len(consumed.data['contacts']))

        consumed = self.client.post(
            '/v1/contacts/queues/{}/consume/?status=pending'.format(q.pk))

        self.assertEqual(0, len(consumed.data['contacts']))

        consumed = self.client.post(
            '/v1/contacts/queues/{}/consume/?status=consumed'.format(q.pk))

        self.assertEqual(2, len(consumed.data['contacts']))


@override_settings(
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    CELERY_ALWAYS_EAGER=True,
    BROKER_BACKEND='memory')
class ContactListTestCase(TestCase):
    fixtures = ['generic_policies']

    def setUp(self):
        self.user = UserFactory()
        # Remove subscription email
        mail.outbox = []
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

    def test_single_policy_subscription(self):
        contact_list = ContactList(author=self.user)
        contact_list.save()

        policy_attribution = ContactListPolicyAttribution(
            contact_list=contact_list,
            policy=ContactListPolicy.objects.get(name='DoubleOptIn'))
        policy_attribution.save()

        contact = Contact(
            contact_list=contact_list,
            address='nope@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        self.assertEqual(Contact.PENDING, contact.status)
        self.assertEqual(1, len(mail.outbox))

    @unittest.skip('Disabled temporarily to allow deploy. FIXME !!!')
    def test_bounce_backend(self):
        contact_list = ContactList(author=self.user)
        contact_list.save()

        policy_attribution = ContactListPolicyAttribution(
            contact_list=contact_list,
            policy=ContactListPolicy.objects.get(name='BounceCheck'))
        policy_attribution.save()

        contact = Contact(
            contact_list=contact_list,
            address='nope@nope', subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        self.assertEqual(1, len(mail.outbox))

        bounce_address = mail.outbox.pop().message().get('Return-Path')

        # sending bounce task
        tasks.handle_bounce(
            {k: v.format(bounce=bounce_address) for k, v in bounce.items()})

        contact.refresh_from_db()
        self.assertEqual(Contact.BOUNCED, contact.status)

        contact = Contact(
            contact_list=contact_list, address='truemail@example.com',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        with fake_time(timezone.now() + timedelta(hours=10)):
            tasks.handle_bounce_expirations()

        contact.refresh_from_db()
        self.assertEqual(Contact.OK, contact.status)

        contact = Contact(
            contact_list=contact_list, address='truemail2@example.com',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        with fake_time(timezone.now() + timedelta(minutes=5)):
            tasks.handle_bounce_expirations()

        contact.refresh_from_db()
        self.assertEqual(
            Contact.PENDING, contact.status,
            'Bounce-check hasn’t expired yet, so the contact should not be OK')

    def test_double_opt_in_backend(self):
        contact_list = ContactList(author=self.user)
        contact_list.save()

        cl_queue_policy = ContactListPolicyAttribution(
            contact_list=contact_list,
            policy=ContactListPolicy.objects.get(name='DoubleOptIn'))
        cl_queue_policy.save()

        contact = Contact(
            contact_list=contact_list,
            address='nope@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        self.assertEqual(1, len(mail.outbox))
        confirmation_msg = mail.outbox.pop().body
        url = URL_REGEX.search(confirmation_msg).group().replace(
            settings.APPLICATION_URL, '')
        response = self.client.get(url)
        self.assertEqual(200, response.status_code)

        contact.refresh_from_db()
        self.assertEqual(Contact.OK, contact.status)

        contact = Contact(
            contact_list=contact_list,
            address='nope2@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        with fake_time(timezone.now() + timedelta(days=8)):
            tasks.handle_opt_ins_expirations()

        contact.refresh_from_db()
        self.assertEqual(Contact.EXPIRED, contact.status)

        contact = Contact(
            contact_list=contact_list,
            address='nope3@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        with fake_time(timezone.now() + timedelta(days=2)):
            tasks.handle_opt_ins_expirations()

        contact.refresh_from_db()
        self.assertEqual(Contact.PENDING, contact.status)

    @unittest.skip('Disabled temporarily to allow deploy. FIXME !!!')
    def test_multiple_backends(self):
        contact_list = ContactList(author=self.user)
        contact_list.save()

        list_attribution_1 = ContactListPolicyAttribution(
            contact_list=contact_list,
            policy=ContactListPolicy.objects.get(name='DoubleOptIn'))
        list_attribution_1.save()

        list_attribution_2 = ContactListPolicyAttribution(
            contact_list=contact_list,
            policy=ContactListPolicy.objects.get(name='BounceCheck'))
        list_attribution_2.save()

        contact = Contact(
            contact_list=contact_list,
            address='nope1@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        self.assertEqual(Contact.PENDING, contact.status)
        self.assertEqual(
            1, len(mail.outbox),
            'Using DoubleOptIn and BounceCheck should only send 1 mail')

        bounce_address = mail.outbox.pop().message().get('Return-Path')

        # sending bounce task
        tasks.handle_bounce({k: v.format(bounce=bounce_address)
                             for k, v in bounce.items()})

        contact.refresh_from_db()
        self.assertEqual(Contact.BOUNCED, contact.status)

        contact = Contact(
            contact_list=contact_list,
            address='nope2@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        confirmation_msg = mail.outbox.pop().body
        url = URL_REGEX.search(confirmation_msg).group().replace(
            settings.APPLICATION_URL, '')
        self.client.get(url)

        contact.refresh_from_db()
        self.assertEqual(
            Contact.OK, contact.status,
            'Contact should be validated if confirmation link is '
            'clicked, even if the bounce hasn’t expired')

        contact = Contact(
            contact_list=contact_list,
            address='nope3@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        with fake_time(timezone.now() + timedelta(hours=5)):
            tasks.handle_bounce_expirations()

        contact.refresh_from_db()
        self.assertEqual(
            Contact.PENDING, contact.status,
            'Bounce expiration should not set contact status to '
            '“ok” if DoubleOptIn is set')

    @unittest.skip('Disabled temporarily to allow deploy. FIXME !!!')
    def test_failed_expiration(self):
        contact_list = ContactList(name="list_1", author=self.user)
        contact_list.save()

        list_attribution = ContactListPolicyAttribution(
            contact_list=contact_list,
            policy=ContactListPolicy.objects.get(name='DoubleOptIn'))
        list_attribution.save()

        contact = Contact(
            contact_list=contact_list,
            address='nope@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        with fake_time(timezone.now() + timedelta(days=8)):
            tasks.handle_opt_ins_expirations()

        with fake_time(timezone.now() + timedelta(days=15)):
            tasks.handle_failed_expirations()

        # Long-time expired contact should be deleted
        self.assertRaises(contact.DoesNotExist, contact.refresh_from_db)

        contact = Contact(
            contact_list=contact_list,
            address='nope@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        contact_list = ContactList(name="list_2", author=self.user)
        contact_list.save()

        list_attribution = ContactListPolicyAttribution(
            contact_list=contact_list,
            policy=ContactListPolicy.objects.get(name='BounceCheck'))
        list_attribution.save()

        contact = Contact(
            contact_list=contact_list,
            address='bounce@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()
        bounce_address = mail.outbox.pop().message().get('Return-Path')

        # sending bounce task
        tasks.handle_bounce(
            {k: v.format(bounce=bounce_address) for k, v in bounce.items()})

        with fake_time(timezone.now() + timedelta(days=15)):
            tasks.handle_failed_expirations()

        # Long-time bounced contact should be deleted
        self.assertRaises(contact.DoesNotExist, contact.refresh_from_db)

        contact = Contact(
            contact_list=contact_list,
            address='ok@nope',
            subscription_ip='127.0.0.1')
        contact.save()
        contact.apply_policies()

        contact.status = Contact.CONSUMED
        contact.save()

        with fake_time(timezone.now() + timedelta(days=15)):
            tasks.handle_consumed_contacts_expirations()

        try:
            contact.refresh_from_db()
        except contact.DoesNotExist:
            self.fail(
                'Long-time consumed contact should not be deleted because '
                'contacts are never consumed compared to collected contacts.')

    def test_http_subscription(self):
        contact_list = ContactList(
            author=self.user,
            contact_fields=[
                {'name': 'name', 'type': 'Char'},
                {'name': 'age', 'type': 'Integer'}])
        contact_list.save()

        response = self.client.post(
            reverse('subscription', kwargs={'uuid': contact_list.uuid}),
            {
                'address': 'yep@example.org',
                'name': 'John Doe',
                'age': 42})
        self.assertEqual(response.status_code, 201)

        try:
            contact = contact_list.contacts.get(address='yep@example.org')
        except Contact.DoesNotExist:
            self.fail('The contact should have been saved')

        self.assertEqual(2, len(contact.properties))

        self.client.post(
            reverse('subscription', kwargs={'uuid': contact_list.uuid}),
            {
                'address': 'nope@example.org',
                'name': 'John Doe',
                'age': 'won’t tell'})

        self.assertRaises(
            Contact.DoesNotExist,
            contact_list.contacts.get, address='nope@example.org')



from django.db.models import Q
from django.contrib.auth.models import Group
from django.contrib.auth import authenticate
from django.utils.translation import ugettext as _
import rest_framework_jwt.settings
from rest_framework import serializers
from rest_framework.serializers import ValidationError
from rest_framework.exceptions import AuthenticationFailed
from rest_framework_jwt.compat import PasswordField
from rest_framework_jwt.serializers import JSONWebTokenSerializer

from munch.core.utils.serializers import HALLinksField
from munch.core.utils.relations import ScopedHyperLinkedRelatedField

from ...utils import UserDisabledException
from ...models import MunchUser
from ...models import Organization
from ...models import APIApplication
from ...models import SmtpApplication
from ...models import OrganizationSettings


class MeSerializer(serializers.HyperlinkedModelSerializer):
    groups = serializers.SlugRelatedField(
        slug_field='name', many=True,
        queryset=Group.objects.all(), required=False)

    _links = HALLinksField(
        nested_endpoints=['regen_secret', 'change_password'],
        view_name='users:munchuser-detail')

    class Meta:
        model = MunchUser
        fields = [
            'identifier', 'first_name', 'last_name', 'secret', 'groups',
            'organization', '_links']
        extra_kwargs = {
            'organization': {'view_name': 'users:organization-detail'}}

    def validate(self, attrs):
        super().validate()
        for field in ('first_name', 'last_name'):
            if field in attrs and attrs[field] == '':
                raise ValidationError(_(
                    '{} cannot be empty'.format(field)))


class APIApplicationSerializer(serializers.HyperlinkedModelSerializer):
    _links = HALLinksField(
        nested_endpoints=['regen_secret'],
        view_name='users:applications-api-detail')

    class Meta:
        model = APIApplication
        fields = ['url', 'identifier', 'secret', '_links']
        read_only_fields = ['secret']
        extra_kwargs = {'url': {'view_name': 'users:applications-api-detail'}}

    def validate(self, attrs):
        identifier = attrs.get('identifier')
        try:
            application = APIApplication.objects.get(
                author=self.context.get('request').user, identifier=identifier)
        except APIApplication.DoesNotExist:
            return attrs
        if self.instance and application.id == self.instance.id:
            return attrs
        else:
            raise ValidationError(_(
                'You already own an api application with this identifier'))


class SmtpApplicationSerializer(serializers.HyperlinkedModelSerializer):
    _links = HALLinksField(
        nested_endpoints=['regen_credentials'],
        view_name='users:applications-smtp-detail')

    class Meta:
        model = SmtpApplication
        fields = ['url', 'username', 'secret', 'identifier', '_links']
        read_only_fields = ['username', 'secret']
        extra_kwargs = {'url': {'view_name': 'users:applications-smtp-detail'}}

    def validate(self, attrs):
        identifier = attrs.get('identifier')
        try:
            application = SmtpApplication.objects.get(
                author=self.context.get('request').user, identifier=identifier)
        except SmtpApplication.DoesNotExist:
            return attrs
        if self.instance and application.id == self.instance.id:
            return attrs
        else:
            raise ValidationError(_(
                'You already own an smtp application with this identifier'))


class MunchUserSerializer(serializers.HyperlinkedModelSerializer):
    api_key = serializers.ReadOnlyField(source='secret')
    groups = serializers.SlugRelatedField(
        slug_field='name', many=True,
        queryset=Group.objects.all(), required=False)

    _links = HALLinksField(
        nested_endpoints=['regen_secret', 'change_password'],
        view_name='users:munchuser-detail')

    class Meta:
        model = MunchUser
        fields = [
            'url', 'identifier', 'first_name', 'last_name', 'groups',
            'api_key', '_links']
        extra_kwargs = {'url': {'view_name': 'users:munchuser-detail'}}


class OrganizationSettingsSerializer(serializers.ModelSerializer):

    class Meta:
        model = OrganizationSettings
        fields = [
            'nickname', 'notify_message_status',
            'notify_optouts', 'external_optout_message']


class ChildOrganizationSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Organization
        fields = ['url', 'name']


class OrganizationSerializer(serializers.HyperlinkedModelSerializer):
    can_attach_files = serializers.ReadOnlyField()
    can_external_optout = serializers.ReadOnlyField()
    name = serializers.CharField(required=True)
    contact_email = serializers.EmailField(required=True)
    settings = OrganizationSettingsSerializer(required=False)
    parent = ScopedHyperLinkedRelatedField(
        required=False, allow_null=True,
        view_name='users:organization-detail',
        queryset=Organization.objects.all())
    _links = HALLinksField(
        nested_endpoints=[
            'opt_outs', 'children', 'invite_user'],
        view_name='users:organization-detail')

    class Meta:
        model = Organization
        fields = [
            'url', 'name', 'contact_email', 'settings', 'parent',
            'can_attach_files', 'can_external_optout', '_links']
        read_only_fields = ['can_attach_files', 'can_external_optout']
        extra_kwargs = {
            'url': {'view_name': 'users:organization-detail'},
            'parent': {'view_name': 'users:organization-detail'}}

    def validate(self, data):
        if not self.instance and not data.get('parent'):
            raise ValidationError({'parent': 'This field may not be blank.'})
        if self.context.get('request').user.organization != data.get('parent'):
            raise ValidationError(
                {'parent': 'Parent organization must be your organization.'})
        return super().validate(data)

    def create(self, validated_data):
        settings_data = validated_data.pop('settings', {})
        organization = Organization(**validated_data)
        organization.clean()
        organization.save()

        if settings_data:
            for field, value in settings_data.items():
                setattr(organization.settings, field, value)
            organization.settings.save()
        return organization

    def update(self, instance, validated_data):
        settings_data = validated_data.pop('settings', {})
        if settings_data:
            for field, value in settings_data.items():
                setattr(instance.settings, field, value)
            instance.settings.save()
        return super().update(instance, validated_data)


class PasswordResetInitSerializer(serializers.Serializer):
    email = serializers.EmailField()


class PasswordChangeSerializer(serializers.Serializer):
    old_password = serializers.CharField(style={'input_type': 'password'})
    new_password = serializers.CharField(style={'input_type': 'password'})

    def __init__(self, user=None, *args, **kwargs):
        self.user = user
        super().__init__(*args, **kwargs)

    def validate_old_password(self, value):
        if self.user.check_password(value):
            return value
        raise ValidationError(
            _('Your old password is incorrect.'))


class MunchJSONWebTokenSerializer(JSONWebTokenSerializer):
    def __init__(self, *args, **kwargs):
        """ Dynamically add the USERNAME_FIELD to self.fields. """
        super().__init__(*args, **kwargs)

        self.fields[self.username_field] = serializers.CharField(
            required=False)
        self.fields['password'] = PasswordField(
            write_only=True, required=False)

    def validate(self, attrs):
        from django.contrib.auth.models import update_last_login

        username = attrs.get(self.username_field)
        password = attrs.get('password')
        # API Key authentication
        if username == 'api' and password:
            user = authenticate(secret=password)
            if user is None:
                raise AuthenticationFailed('Invalid API key')
            if not user.is_active:
                raise UserDisabledException()
            payload = rest_framework_jwt.settings.api_settings.\
                JWT_PAYLOAD_HANDLER(user)
            update_last_login(None, user)
        # Username/Password authentication
        elif all([username, password]):
            user = authenticate(username=username, password=password)
            if user:
                if not user.is_active:
                    raise UserDisabledException()

                payload = rest_framework_jwt.settings.api_settings.\
                    JWT_PAYLOAD_HANDLER(user)
                update_last_login(None, user)
            else:
                msg = _('Unable to login with provided credentials.')
                raise AuthenticationFailed(msg)
        else:
            msg = _(
                'Must include "{username_field}" and "password".')
            msg = msg.format(username_field=self.username_field)
            raise ValidationError(msg)

        return {
            'token': rest_framework_jwt.settings.api_settings.
            JWT_ENCODE_HANDLER(payload),
            'user': user}


class InvitationSerializer(serializers.ModelSerializer):
    class Meta:
        model = MunchUser
        fields = ['identifier']

    def validate(self, attrs):
        user = MunchUser(**attrs)
        request = self.context.get('request')
        organization_id = self.context.get('organization_id')
        if not organization_id or not organization_id.isdigit() or \
                int(organization_id) not in Organization.objects.filter(
                    Q(id__in=[request.user.organization_id]) |
                    Q(parent_id=request.user.organization_id)).values_list(
                        'id', flat=True):
            raise ValidationError(_('Invalid organization.'))
        user.organization_id = self.context.get('organization_id')
        user.clean()
        return attrs



from rest_framework import serializers

from munch.core.utils.serializers import HALLinksField

from ...models import Mail
from ...models import MailBatch
from ...models import MailStatus


class NestedTrackingSummarySerializer(serializers.Serializer):
    opened = serializers.DateTimeField(source='first_open')
    open_time = serializers.IntegerField()
    clicked = serializers.ListField(
        source='clicks', child=serializers.DictField())


class MailStatusSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = MailStatus
        fields = ['status', 'creation_date', 'status_code', 'raw_msg']


class MailSerializer(serializers.HyperlinkedModelSerializer):
    last_status = MailStatusSerializer(read_only=True)
    tracking = NestedTrackingSummarySerializer(
        read_only=True, source='tracking_info')
    _links = HALLinksField(
        nested_endpoints=['statuses'],
        view_name='transactional:mail-detail')

    class Meta:
        model = Mail
        fields = [
            'url', 'identifier', 'sender', 'creation_date', 'last_status',
            'tracking', 'batch', 'track_open', 'track_clicks', '_links']
        extra_kwargs = {
            'url': {'view_name': 'transactional:mail-detail'},
            'batch': {'view_name': 'transactional:batch-detail'}}


class MailBatchSerializer(serializers.HyperlinkedModelSerializer):
    _links = HALLinksField(
        nested_endpoints=['mails', 'stats', 'bounces', 'opt_outs'],
        view_name='transactional:batch-detail')

    class Meta:
        model = MailBatch
        fields = ['url', 'creation_date', 'name', 'category', '_links']
        extra_kwargs = {
            'url': {'view_name': 'transactional:batch-detail'},
            'category': {'view_name': 'core:category-detail'}}



from rest_framework import mixins
from rest_framework import viewsets

from munch.core.utils.views import OrganizationOwnedViewSetMixin

from ...models import OptOut
from .filters import OptOutFilter
from .serializers import OptOutSerializer


class OptOutViewSet(
        OrganizationOwnedViewSetMixin, mixins.CreateModelMixin,
        viewsets.ReadOnlyModelViewSet):
    model = OptOut
    serializer_class = OptOutSerializer
    filter_class = OptOutFilter



# -*- coding: utf-8 -*-
from django.db import migrations
from django.core.management.sql import emit_post_migrate_signal

PERMISSIONS = {
    'mailstatus': [
        ('add_mailstatus', 'Can add mail status'),
        ('change_mailstatus', 'Can change mail status'),
        ('change_mine_mailstatus', 'Can change_mine mail status'),
        ('change_organizations_mailstatus', 'Can change_organizations mail status'),
        ('delete_mailstatus', 'Can delete mail status'),
        ('delete_mine_mailstatus', 'Can delete_mine mail status'),
        ('delete_organizations_mailstatus', 'Can delete_organizations mail status'),
        ('view_mailstatus', 'Can view mail status'),
        ('view_mine_mailstatus', 'Can view_mine mail status'),
        ('view_organizations_mailstatus', 'Can view_organizations mail status'), ],
    'mail': [
        ('add_mail', 'Can add mail'),
        ('change_mail', 'Can change mail'),
        ('change_mine_mail', 'Can change_mine mail'),
        ('change_organizations_mail', 'Can change_organizations mail'),
        ('delete_mail', 'Can delete mail'),
        ('delete_mine_mail', 'Can delete_mine mail'),
        ('delete_organizations_mail', 'Can delete_organizations mail'),
        ('view_mail', 'Can view mail'),
        ('view_mine_mail', 'Can view_mine mail'),
        ('view_organizations_mail', 'Can view_organizations mail'), ],
    'message': [
        ('add_message', 'Can add Message'),
        ('change_message', 'Can change Message'),
        ('change_mine_message', 'Can change_mine Message'),
        ('change_organizations_message', 'Can change_organizations Message'),
        ('delete_message', 'Can delete Message'),
        ('delete_mine_message', 'Can delete_mine Message'),
        ('delete_organizations_message', 'Can delete_organizations Message'),
        ('previewsend_message', 'Can previewsend Message'),
        ('previewsend_mine_message', 'Can previewsend_mine Message'),
        ('previewsend_organizations_message', 'Can previewsend_organizations Message'),
        ('view_message', 'Can view Message'),
        ('view_mine_message', 'Can view_mine Message'),
        ('view_organizations_message', 'Can view_organizations Message'), ],
    'messageattachment': [
        ('add_messageattachment', 'Can add message attachment'),
        ('change_messageattachment', 'Can change message attachment'),
        ('change_mine_messageattachment', 'Can change_mine message attachment'),
        ('change_organizations_messageattachment', 'Can change_organizations message attachment'),
        ('delete_messageattachment', 'Can delete message attachment'),
        ('delete_mine_messageattachment', 'Can delete_mine message attachment'),
        ('delete_organizations_messageattachment', 'Can delete_organizations message attachment'),
        ('view_messageattachment', 'Can view message attachment'),
        ('view_mine_messageattachment', 'Can view_mine message attachment'),
        ('view_organizations_messageattachment', 'Can view_organizations message attachment'), ],
    'previewmail': [
        ('add_previewmail', 'Can add preview mail'),
        ('change_mine_previewmail', 'Can change_mine preview mail'),
        ('change_organizations_previewmail', 'Can change_organizations preview mail'),
        ('change_previewmail', 'Can change preview mail'),
        ('delete_mine_previewmail', 'Can delete_mine preview mail'),
        ('delete_organizations_previewmail', 'Can delete_organizations preview mail'),
        ('delete_previewmail', 'Can delete preview mail'),
        ('view_mine_previewmail', 'Can view_mine preview mail'),
        ('view_organizations_previewmail', 'Can view_organizations preview mail'),
        ('view_previewmail', 'Can view preview mail'), ],
}
GROUP_PERMISSIONS = {
    'administrators': {
        'mailstatus': [
            'view_organizations_mailstatus', ],
        'mail': [
            'add_mail',
            'change_organizations_mail',
            'delete_organizations_mail',
            'view_organizations_mail', ],
        'message': [
            'add_message',
            'change_organizations_message',
            'delete_organizations_message',
            'previewsend_organizations_message',
            'view_organizations_message', ],
        'messageattachment': [
            'add_messageattachment',
            'change_organizations_messageattachment',
            'delete_organizations_messageattachment',
            'view_organizations_messageattachment', ],
        'previewmail': [
            'change_organizations_previewmail',
            'delete_organizations_previewmail',
            'view_organizations_previewmail', ],
    },
    'managers': {
        'mailstatus': [
            'view_organizations_mailstatus', ],
        'mail': [
            'add_mail',
            'change_organizations_mail',
            'delete_organizations_mail',
            'view_organizations_mail', ],
        'message': [
            'add_message',
            'change_organizations_message',
            'delete_organizations_message',
            'previewsend_organizations_message',
            'view_organizations_message', ],
        'messageattachment': [
            'add_messageattachment',
            'change_organizations_messageattachment',
            'delete_organizations_messageattachment',
            'view_organizations_messageattachment', ],
        'previewmail': [
            'change_organizations_previewmail',
            'delete_organizations_previewmail',
            'view_organizations_previewmail', ],
    },
    'users': {
        'mailstatus': [
            'view_mine_mailstatus', ],
        'mail': [
            'add_mail',
            'change_mine_mail',
            'delete_mine_mail',
            'view_mine_mail', ],
        'message': [
            'add_message',
            'change_mine_message',
            'delete_mine_message',
            'previewsend_mine_message',
            'view_mine_message', ],
        'messageattachment': [
            'add_messageattachment',
            'change_mine_messageattachment',
            'delete_mine_messageattachment',
            'view_mine_messageattachment', ],
        'previewmail': [
            'view_mine_previewmail', ],
    },
    'collaborators': {
        'mailstatus': [
            'view_organizations_mailstatus', ],
        'mail': [
            'add_mail',
            'change_organizations_mail',
            'delete_organizations_mail',
            'view_organizations_mail', ],
        'message': [
            'add_message',
            'change_organizations_message',
            'delete_organizations_message',
            'previewsend_organizations_message',
            'view_organizations_message', ],
        'messageattachment': [
            'add_messageattachment',
            'change_organizations_messageattachment',
            'delete_organizations_messageattachment',
            'view_organizations_messageattachment', ],
        'previewmail': [
            'change_organizations_previewmail',
            'delete_organizations_previewmail',
            'view_organizations_previewmail', ],
    },
}


def update_content_types(apps, schema_editor):
    db_alias = schema_editor.connection.alias
    emit_post_migrate_signal(False, 'default', db_alias)


def load_permissions(apps, schema_editor):
    Group = apps.get_model('auth', 'group')
    Permission = apps.get_model('auth', 'permission')
    ContentType = apps.get_model('contenttypes', 'contenttype')

    # Delete previous permissions
    for model in PERMISSIONS:
        content_type = ContentType.objects.get(
            app_label='campaigns', model=model)
        Permission.objects.filter(content_type=content_type).delete()

    # Load permissions
    for model_name, permissions in PERMISSIONS.items():
        for permission_codename, permission_name in permissions:
            content_type = ContentType.objects.get(
                app_label='campaigns', model=model_name)
            if not Permission.objects.filter(
                    codename=permission_codename,
                    content_type=content_type).exists():
                Permission.objects.create(
                    name=permission_name,
                    codename=permission_codename,
                    content_type=content_type)

    # Group permissions
    for group_name, models in GROUP_PERMISSIONS.items():
        group, _ = Group.objects.get_or_create(name=group_name)
        for model_name, permissions in models.items():
            content_type = ContentType.objects.get(
                app_label='campaigns', model=model_name)
            for codename in permissions:
                permission, _ = Permission.objects.get_or_create(
                    codename=codename, content_type_id=content_type.id)
                group.permissions.add(permission)
                group.save()


class Migration(migrations.Migration):
    # TODO: Add your dependencies here:
    #   dependencies = [('campaigns', '0001_initial')]
    dependencies = [('campaigns', '0001_initial')]
    operations = [
        migrations.RunPython(update_content_types, reverse_code=None),
        migrations.RunPython(load_permissions, reverse_code=None)]



from django.conf.urls import url
from django.conf.urls import include
from rest_framework.urlpatterns import format_suffix_patterns

from munch.urls import api_urlpatterns_v1

from . import views

urlpatterns = format_suffix_patterns([
    url(r'^$', views.api_root, name='api-root'),
    url(r'^images/?$', views.ImageCreate.as_view(), name='image-create'),
    url(
        r'^images/(?P<pk>[_\w-]+)/?$',
        views.ImageDetail.as_view(), name='image-detail'),
])

api_urlpatterns_v1 += [
    url(r'^upload-store/', include(urlpatterns, namespace='upload-store'))]



from django.test import TestCase
from rest_framework.test import APIClient
from faker import Factory as FakerFactory

from ...models import MunchUser
from ...models import Organization
from ...models import APIApplication
from ...models import SmtpApplication
from ...tests.factories import UserFactory
from ...tests.factories import OrganizationFactory
from ...tests.factories import APIApplicationFactory
from ...tests.factories import SmtpApplicationFactory

faker = FakerFactory.create()


class APIApplicationAPITestCase(TestCase):
    def setUp(self):
        self.api_prefix = 'v1'
        self.user = UserFactory(groups=['managers'])
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

    def test_create_api_application(self):
        response = self.client.post(
            '/{}/applications/api/'.format(self.api_prefix),
            {'identifier': 'api-foo'}, format='json')

        self.assertEqual(response.status_code, 201)
        self.assertEqual(APIApplication.objects.count(), 1)

    def test_read_only_secret(self):
        response = self.client.post(
            '/{}/applications/api/'.format(self.api_prefix),
            {'identifier': 'api-foo', 'secret': '1'}, format='json')
        self.assertNotEqual(response.data['secret'], '1')

        response = self.client.patch(
            response.data['url'], {'secret': '1'}, format='json')
        self.assertNotEqual(response.data['secret'], '1')

    def test_regen_secret(self):
        response = self.client.post(
            '/{}/applications/api/'.format(self.api_prefix),
            {'identifier': 'api-foo'}, format='json')

        self.assertEqual(response.data['identifier'], 'api-foo')
        self.assertTrue(response.data['secret'])

        secret = response.data['secret']
        application = APIApplication.objects.all().first()

        response = self.client.post(
            '/{}/applications/api/{}/regen_secret/'.format(
                self.api_prefix, application.pk))

        self.assertEqual(response.status_code, 200)
        self.assertNotEqual(response.data, secret)

    def test_unique_identifier(self):
        response = self.client.post(
            '/{}/applications/api/'.format(self.api_prefix),
            {'identifier': 'api-foo'}, format='json')

        response = self.client.post(
            '/{}/applications/api/'.format(self.api_prefix),
            {'identifier': 'api-foo'}, format='json')

        self.assertEqual(response.status_code, 400)
        self.assertEqual(APIApplication.objects.count(), 1)
        self.assertIn('non_field_errors', response.json())

    def test_update_identifier(self):
        response = self.client.post(
            '/{}/applications/api/'.format(self.api_prefix),
            {'identifier': 'api-foo'}, format='json')

        response = self.client.put(
            response.json().get('url'),
            {'identifier': 'api-foo-updated'}, format='json')

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json().get('identifier'), 'api-foo-updated')

    def test_permissions_not_manager_same_organization(self):
        # Create an API application as manager
        response = self.client.post(
            '/{}/applications/api/'.format(self.api_prefix),
            {'identifier': 'api'}, format='json')
        url = response.json().get('url')

        # Test a simple user
        user = UserFactory(
            groups=['users'], organization=self.user.organization)
        self.client.login(identifier=user.identifier, password='password')

        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        response = self.client.get(
            '/{}/applications/api/'.format(self.api_prefix))
        self.assertEqual(response.status_code, 403)
        response = self.client.post(
            '/{}/applications/api/'.format(self.api_prefix),
            {'identifier': 'api-foo'}, format='json')
        self.assertEqual(response.status_code, 403)

        # Test a collaborator
        collaborator = UserFactory(
            groups=['collaborators'], organization=self.user.organization)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        self.client.login(
            identifier=collaborator.identifier, password='password')
        response = self.client.get(
            '/{}/applications/api/'.format(self.api_prefix))
        self.assertEqual(response.status_code, 403)
        response = self.client.post(
            '/{}/applications/api/'.format(self.api_prefix),
            {'identifier': 'api-foo'}, format='json')
        self.assertEqual(response.status_code, 403)

    def test_permissions_another_organization(self):
        # Create an API application with first organization
        response = self.client.post(
            '/{}/applications/api/'.format(self.api_prefix),
            {'identifier': 'api-foo'}, format='json')
        url = response.json().get('url')

        # Create another manager for another organization
        another_manager = UserFactory(groups=['managers'])
        self.client.login(
            identifier=another_manager.identifier, password='password')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)

        response = self.client.get(
            '/{}/applications/api/'.format(self.api_prefix))
        self.assertEqual(response.json().get('count'), 0)


class SmtpApplicationAPITestCase(TestCase):
    def setUp(self):
        self.api_prefix = 'v1'
        self.user = UserFactory(groups=['managers'])
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

    def test_create_smtp_application(self):
        response = self.client.post(
            '/{}/applications/smtp/'.format(self.api_prefix),
            {'identifier': 'smtp-foo'}, format='json')

        self.assertEqual(response.status_code, 201)
        self.assertEqual(SmtpApplication.objects.count(), 1)
        self.assertTrue(response.json().get('username'))
        self.assertTrue(response.json().get('secret'))

    def test_read_only_credentials(self):
        response = self.client.post(
            '/{}/applications/smtp/'.format(self.api_prefix),
            {'identifier': 'smtp-foo', 'secret': '1', 'username': 'jambon'},
            format='json')
        self.assertNotEqual(response.data['secret'], '1')
        self.assertNotEqual(response.data['username'], 'jambon')

        response = self.client.patch(
            response.data['url'],
            {'secret': '1', 'username': 'jambon'}, format='json')
        self.assertNotEqual(response.data['secret'], '1')
        self.assertNotEqual(response.data['username'], 'jambon')

    def test_regen_credentials(self):
        response = self.client.post(
            '/{}/applications/smtp/'.format(self.api_prefix),
            {'identifier': 'smtp-foo'}, format='json')

        self.assertEqual(response.data['identifier'], 'smtp-foo')
        self.assertTrue(response.data['secret'])
        self.assertTrue(response.data['username'])

        secret = response.data['secret']
        username = response.data['username']

        application = SmtpApplication.objects.all().first()

        response = self.client.post(
            '/{}/applications/smtp/{}/regen_credentials/'.format(
                self.api_prefix, application.pk))

        self.assertEqual(response.status_code, 200)
        self.assertNotEqual(response.json().get('secret'), secret)
        self.assertNotEqual(response.json().get('username'), username)

    def test_unique_identifier(self):
        response = self.client.post(
            '/{}/applications/smtp/'.format(self.api_prefix),
            {'identifier': 'smtp-foo'}, format='json')

        response = self.client.post(
            '/{}/applications/smtp/'.format(self.api_prefix),
            {'identifier': 'smtp-foo'}, format='json')

        self.assertEqual(response.status_code, 400)
        self.assertEqual(SmtpApplication.objects.count(), 1)
        self.assertIn('non_field_errors', response.json())

    def test_update_identifier(self):
        response = self.client.post(
            '/{}/applications/smtp/'.format(self.api_prefix),
            {'identifier': 'smtp-foo'}, format='json')

        response = self.client.put(
            response.json().get('url'),
            {'identifier': 'smtp-foo-updated'}, format='json')

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json().get('identifier'), 'smtp-foo-updated')

    def test_permissions_not_manager_same_organization(self):
        # Create an API application as manager
        response = self.client.post(
            '/{}/applications/smtp/'.format(self.api_prefix),
            {'identifier': 'smtp'}, format='json')
        url = response.json().get('url')

        # Test a simple user
        user = UserFactory(
            groups=['users'], organization=self.user.organization)
        self.client.login(identifier=user.identifier, password='password')

        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        response = self.client.get(
            '/{}/applications/smtp/'.format(self.api_prefix))
        self.assertEqual(response.status_code, 403)
        response = self.client.post(
            '/{}/applications/smtp/'.format(self.api_prefix),
            {'identifier': 'smtp-foo'}, format='json')
        self.assertEqual(response.status_code, 403)

        # Test a collaborator
        collaborator = UserFactory(
            groups=['collaborators'], organization=self.user.organization)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        self.client.login(
            identifier=collaborator.identifier, password='password')
        response = self.client.get(
            '/{}/applications/smtp/'.format(self.api_prefix))
        self.assertEqual(response.status_code, 403)
        response = self.client.post(
            '/{}/applications/smtp/'.format(self.api_prefix),
            {'identifier': 'smtp-foo'}, format='json')
        self.assertEqual(response.status_code, 403)

    def test_permissions_another_organization(self):
        # Create an API application with first organization
        response = self.client.post(
            '/{}/applications/smtp/'.format(self.api_prefix),
            {'identifier': 'smtp-foo'}, format='json')
        url = response.json().get('url')

        # Create another manager for another organization
        another_manager = UserFactory(groups=['managers'])
        self.client.login(
            identifier=another_manager.identifier, password='password')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)

        response = self.client.get(
            '/{}/applications/smtp/'.format(self.api_prefix))
        self.assertEqual(response.json().get('count'), 0)


class OrganizationAPITestCase(TestCase):
    def setUp(self):
        self.api_prefix = 'v1'
        self.administrator = UserFactory(groups=['administrators'])
        self.manager = UserFactory(
            groups=['managers'], organization=self.administrator.organization)
        self.user = UserFactory(
            groups=['users'], organization=self.administrator.organization)

        self.client = APIClient()
        self.client.login(
            identifier=self.administrator.identifier, password='password')

        self.url = '/{}/organizations/'.format(self.api_prefix)

    def test_can_only_see_mine_organization(self):
        user = UserFactory(groups=['administrators'])
        self.assertEqual(Organization.objects.count(), 2)

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['results']), 1)

        self.client.login(
            identifier=self.manager.identifier, password='password')
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['results']), 1)

        self.client.login(
            identifier=self.user.identifier, password='password')
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['results']), 1)

        self.client.login(identifier=user.identifier, password='password')
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['results']), 1)

    def test_nobody_can_create_non_child_organization(self):
        data = {'name': 'name', 'contact_email': 'contact@example.com' }

        response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(Organization.objects.count(), 1)

        self.client.login(
            identifier=self.manager.identifier, password='password')
        response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(Organization.objects.count(), 1)

        self.client.login(
            identifier=self.user.identifier, password='password')
        response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(Organization.objects.count(), 1)

    def test_only_administrator_can_create_child_organization(self):
        data = {
            'name': 'name',
            'parent': self.client.get(self.url).json()['results'][0]['url'],
            'contact_email': 'contact@example.com' }

        # Manager
        self.client.login(
            identifier=self.manager.identifier, password='password')
        response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(Organization.objects.count(), 1)

        # User
        self.client.login(
            identifier=self.user.identifier, password='password')
        response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(Organization.objects.count(), 1)

        # Administrator
        self.client.login(
            identifier=self.administrator.identifier, password='password')
        response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Organization.objects.count(), 2)

    def test_cannot_create_child_of_not_owned_organization(self):
        user = UserFactory(groups=['administrators'])
        self.assertEqual(Organization.objects.count(), 2)

        self.client.login(identifier=user.identifier, password='password')
        data = {
            'name': 'name',
            'parent': self.client.get(self.url).json()['results'][0]['url'],
            'contact_email': 'contact@example.com' }

        self.client.login(
            identifier=self.administrator.identifier, password='password')
        response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(Organization.objects.count(), 2)
        self.assertIn('parent', response.json())

    def test_cannot_add_child_on_child_organization(self):
        data = {
            'name': 'name',
            'parent': self.client.get(self.url).json()['results'][0]['url'],
            'contact_email': 'contact@example.com' }

        response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Organization.objects.count(), 2)

        data = {
            'name': 'name',
            'parent': response.json()['url'],
            'contact_email': 'contact@example.com' }

        response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 400)
        self.assertIn('parent', response.json())
        self.assertEqual(Organization.objects.count(), 2)

    def test_only_adminstrator_can_see_children_organizations(self):
        data = {
            'name': 'name',
            'parent': self.client.get(self.url).json()['results'][0]['url'],
            'contact_email': 'contact@example.com' }

        response = self.client.post(self.url, data)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Organization.objects.count(), 2)

        organization = self.client.get(self.url).json()['results'][0]
        response = self.client.get(organization['_links']['children']['href'])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)

        response = self.client.get(
            response.json()[0]['_links']['children']['href'])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 0)


class InvitationAPITestCase(TestCase):
    def setUp(self):
        self.api_prefix = 'v1'

        self.administrator = UserFactory(groups=['administrators'])
        self.user = UserFactory(
            groups=['users'], organization=self.administrator.organization)

        self.client = APIClient()
        self.client.login(
            identifier=self.administrator.identifier, password='password')

    def test_simple_invitation(self):
        organization_url = '/{}/organizations/{}/'.format(
            self.api_prefix, self.administrator.organization_id)
        invite_user_url = self.client.get(organization_url).json()[
            '_links']['invite_user']['href']
        data = {"identifier": faker.safe_email()}
        resp = self.client.post(invite_user_url, data)
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp.json()["identifier"], data['identifier'])
        self.assertEqual(self.administrator.organization.users.count(), 3)
        self.assertFalse(MunchUser.objects.filter(
            identifier=data['identifier']).first().is_active)

    def test_already_used_identifier(self):
        organization_url = '/{}/organizations/{}/'.format(
            self.api_prefix, self.administrator.organization_id)
        invite_user_url = self.client.get(organization_url).json()[
            '_links']['invite_user']['href']
        data = {"identifier": self.user.identifier}
        resp = self.client.post(invite_user_url, data)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(self.administrator.organization.users.count(), 2)

    def test_required_fields(self):
        organization_url = '/{}/organizations/{}/'.format(
            self.api_prefix, self.administrator.organization_id)
        invite_user_url = self.client.get(organization_url).json()[
            '_links']['invite_user']['href']
        data = {"identifier": ""}
        resp = self.client.post(invite_user_url, data)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(self.administrator.organization.users.count(), 2)

    def test_children_organization(self):
        child = OrganizationFactory(parent=self.administrator.organization)
        child_url = '/{}/organizations/{}/'.format(self.api_prefix, child.id)

        invite_user_url = self.client.get(child_url).json()[
            '_links']['invite_user']['href']

        data = {"identifier": faker.safe_email()}
        resp = self.client.post(invite_user_url, data)
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp.json()["identifier"], data['identifier'])
        self.assertEqual(self.administrator.organization.users.count(), 2)
        self.assertEqual(child.users.count(), 1)
        self.assertFalse(MunchUser.objects.filter(
            identifier=data['identifier']).first().is_active)

    def test_not_owned_organization(self):
        another_org = OrganizationFactory()
        another_org_url = '/{}/organizations/{}/invite_user/'.format(
            self.api_prefix, another_org.id)

        data = {"identifier": faker.safe_email()}
        resp = self.client.post(another_org_url, data)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(self.administrator.organization.users.count(), 2)
        self.assertEqual(another_org.users.count(), 0)



from unittest.mock import patch

import django_fsm
from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings
from django.core.exceptions import ValidationError

from munch.core.utils.tests import temporary_settings
from munch.apps.spamcheck import SpamResult
from munch.apps.users.tests.factories import UserFactory
from munch.apps.spamcheck.tests import get_spam_result_mock

from ..exceptions import InvalidSubmitedData
from .factories import MessageFactory


@override_settings(SKIP_SPAM_CHECK=True, BYPASS_DNS_CHECKS=True)
class MessageTests(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

    def test_empty_html(self):
        self.message.html = ' '
        with self.assertRaises(ValidationError):
            self.message.save()

    def test_add_messages_update_status(self):
        self.message.save()
        self.assertEqual(self.message.status, self.message.MSG_OK)

    def test_unauthorized_external_optout(self):
        self.assertEqual(
            self.message.author.organization.can_external_optout, False)
        self.message.external_optout = True

        with self.assertRaises(ValidationError):
            self.message.full_clean()
            self.message.save()

    def test_authorized_external_optout(self):
        self.message.author.organization.can_external_optout = True
        self.message.author.organization.save()

        self.message.external_optout = True
        self.message.full_clean()
        self.message.save()


@override_settings(SKIP_SPAM_CHECK=True)
class MessageStatesTest(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

    def test_allowed_transition(self):
        """ NEW -> CANCELED"""
        self.message.status = self.message.MSG_ISSUES
        self.message.html = '<html>{}</html>'.format(
            settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.message.save()
        self.assertEqual(self.message.status, self.message.MSG_OK)

    def test_unallowed_transition(self):
        """NEW -> SENT"""
        self.message.status = self.message.SENT
        with self.assertRaises(django_fsm.TransitionNotAllowed):
            self.message.save()


@override_settings(SKIP_SPAM_CHECK=True)
class MessageTest(TestCase):
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

    def test_plaintext_render(self):
        self.message.html = '<h1>Foubarr{}</h1>'.format(
            settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertIn('# Foubarr', self.message.mk_plaintext())

    def test_plaintext_render_keeps_accents(self):
        self.message.html = '<h1>é</h1>'
        self.assertIn('é', self.message.mk_plaintext())

    def test_css_inlining(self):
        self.message.html = (
            '<style>h1 {color: red}</style><h1>Foubarré</h1>%s') % (
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertIn('<h1 style="color:red">', self.message.mk_html())

    def test_css_inlining_keep_property_caps(self):
        self.message.html = (
            '<style>h1 {Margin: 1em}</style><h1>Foubarré</h1>%s') % (
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertIn('<h1 style="Margin:1em">', self.message.mk_html())

    def test_unsubscribe_link_preexisting(self):
        self.message.html = '<body><h1>Foubarr</h1><p>{}</p></body>'.format(
            settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertEqual(
            self.message.mk_html().count(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER']), 1)

    def test_clean_message_script(self):
        self.message.html = (
            '<head><script>alert("foo");</script>'
            '</head><body>{}</body>').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertNotIn('<script>', self.message.mk_html())

    def test_clean_message_js(self):
        self.message.html = (
            '<head></head><body><h1 onclick="alert(\'foo\');">'
            '</h1>{}</body>').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertNotIn('alert', self.message.mk_html())

    def test_detach_img(self):
        self.message.html = (
            '<img src="data:image/gif;base64'
            ',R0lGODdhAQABAIABAOvr6////ywAAAAAAQABAAACAkQBADs="/>{}').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertIn('/uploads/images/', self.message.mk_html())

    def test_retain_body_attributes(self):
        self.message.html = (
            '<body style="width: 100%"><h1>A</h1><p>B</p>{}</body>').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertIn('style="width: 100%"', self.message.mk_html())

    def test_do_not_div_wrap_document(self):
        """
        see bug #1542
        """
        self.message.html = (
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" '
            '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
            '<html xmlns="http://www.w3.org/1999/xhtml">'
            '<body leftmargin="0" marginwidth="0" topmargin="0" '
            'marginheight="0" offset="0">{}</body></html>').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertNotIn('<div>', self.message.mk_html())

    def test_preserve_doctype(self):
        self.message.html = (
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" '
            '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
            '<html xmlns="http://www.w3.org/1999/xhtml">{}</html>').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertIn((
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0'
            ' Transitional//EN" "http://www.w3.org/TR/'
            'xhtml1/DTD/xhtml1-transitional.dtd">').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER']),
            self.message.mk_html())
        self.assertIn(
            'xmlns="http://www.w3.org/1999/xhtml', self.message.mk_html())

    def test_strip_comments(self):
        self.message.html = '<body><!-- IM COMMENT -->{}</body>'.format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertNotIn('COMMENT', self.message.mk_html())

    def test_preserve_conditional_comments(self):
        self.message.html = (
            "<body><!--[if IE 6]><p>MARK</p><![endif]-->{}</body>").format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertIn('MARK', self.message.mk_html())

    def test_remove_tbody(self):
        """Checks that <tbody> are removed but their contents remains """
        self.message.html = (
            "<body><table><tbody><tr><td>PLACEHOLDER</td>"
            "</tr></tbody></table>{}</body>").format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertNotIn('<tbody>', self.message.mk_html())
        self.assertIn('<tr><td>PLACEHOLDER</td></tr>', self.message.mk_html())

    def test_remove_link(self):
        self.message.html = (
            '<head><link rel="home" href="/" /></head><body>{}</body>').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertNotIn('<link', self.message.mk_html())

    def test_fetch_img(self):
        self.message.detach_images = True
        self.message.html = (
            '<img src="http://www.oasiswork.fr/uploads'
            '/logo_oasiswork.jpg" />{}').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        html = self.message.mk_html()
        self.assertIn('/uploads/images/', html)
        self.assertNotIn('src="http://www.oasiswork.fr', html)

    def test_do_not_fetch_img(self):
        self.message.detach_images = False
        self.message.html = (
            '<img src="http://www.oasiswork.fr/uploads'
            '/logo_oasiswork.jpg" />{}').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        self.assertIn(
            'src="http://www.oasiswork.fr/uploads/logo_oasiswork.jpg"',
            self.message.mk_html())

    def test_invalid_html(self):
        """ Invalid because of missing src attr
        """
        with self.assertRaises(InvalidSubmitedData):
            self.message.html = '<img'
            self.message.mk_html()

    def test_spam_check(self):
        self.message.html = '<body><h1>Foubarr</h1> {}</body>'.format(
            settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])

        with patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=get_spam_result_mock):
            check_res = self.message.spam_check()
        self.assertIsInstance(check_res, SpamResult)
